# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from multiprocessing.pool import ThreadPool
from typing import Optional

from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QFontMetrics, QKeyEvent, QResizeEvent
from PyQt5.QtWidgets import QGridLayout, QPlainTextEdit, QSizePolicy, QWidget

from . import Chat
from ..commands.eval import eval_f


class SendBox(QPlainTextEdit):
    def __init__(self, area: "SendArea") -> None:
        super().__init__()
        self.area  = area
        self._pool = ThreadPool(1)

        self.setPlaceholderText("Send a message…")
        self.setCenterOnScroll(False)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.document().setDocumentMargin(0)

        self.textChanged.connect(self.updateGeometry)


    def get_box_height(self, lines: Optional[int] = None) -> int:
        widget_margin   = self.contentsMargins()
        document_margin = self.document().documentMargin()
        font_height     = QFontMetrics(self.document().defaultFont()).height()

        height = (widget_margin.top()    +
                  widget_margin.bottom() +
                  document_margin * 2    +
                  # Extra pixel prevents wordwrap-scroll-box-grows from hiding
                  # the first line
                  font_height * (lines or self.document().lineCount()) + 1)

        max_height = self.area.chat.height() // 2

        if lines:
            return min(height, max_height)

        if height >= max_height:
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
            return max_height

        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        return height


    def sizeHint(self) -> QSize:
        return QSize(1, self.get_box_height())

    def minimumSizeHint(self) -> QSize:
        return QSize(1, self.get_box_height(lines=1))


    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)
        self.updateGeometry()  # on main window resize, adjust max height


    def keyPressEvent(self, event: QKeyEvent) -> None:
        if event.key() in (Qt.Key_Return, Qt.Key_Enter):  # Enter = keypad
            if event.modifiers() in (Qt.NoModifier, Qt.KeypadModifier):
                if not event.isAutoRepeat():
                    self.send()
            else:
                self.insertPlainText("\n")  # Make it work with Shift/Alt/Ctrl
            return

        super().keyPressEvent(event)


    def send(self) -> None:
        text = self.toPlainText().rstrip()
        if not text:
            return

        self.clear()

        self._pool.apply_async(
            func           = eval_f,
            args           = (self.area.chat, text),
            error_callback = self.on_error
        )


    @staticmethod
    def on_error(err: BaseException) -> None:
        raise err


class SendArea(QWidget):
    def __init__(self, chat: Chat) -> None:
        super().__init__(chat)
        self.chat = chat

        self.grid = QGridLayout(self)
        self.original_margin = self.grid.contentsMargins()
        self.grid.setContentsMargins(0, 0, 0, 0)

        self.box = SendBox(self)
        self.grid.addWidget(self.box, 0, 0)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import re
from copy import copy
from typing import Any, Callable, ClassVar, Dict, List, Optional, Tuple

from dataclasses import dataclass
from PyQt5.QtCore import QDateTime

from . import main_window, markdown

DATE_FORMAT = "HH:mm:ss"


@dataclass
class Message:
    """Represents a locally generated or received from server message.

    Class variables:
        local_echo_hooks:
            Dict of `identifier: functions` that will receive Message objects
            to [locally echo](https://tinyurl.com/matrix-local-echo)
            when Message.send() is called.
            Identifier can be a string or any other object.

    Attributes:
        room_id:
            ID of the room this message belongs to, required.

        sender_id:
            Matrix user ID of the message's sender, required.

        receiver_id:
            Matrix user ID of the account that received this message.
            If `None` (default), message is assumed to be locally generated.

        markdown:
            Content of the message in plain text/markdown format.
            If not specified, `html` must be and is converted to `markdown`.

        html:
            Content of the message in HTML format.
            If not specified, `markdown` must be and is converted to `html`.

        ms_since_epoch:
            Unix timestamp in microseconds for the message's creation.
            If not specified, the current time will be used.

        avatar_url:
            HTTP(s) URL of the sender's avatar, if he has one."""

    # Can't define a pyqtSignal(this_class) here
    local_echo_hooks: ClassVar[Dict[Any, Callable[["Message"], None]]] = {}

    room_id:     str           = ""
    sender_id:   str           = ""
    receiver_id: Optional[str] = None
    markdown:    str           = ""
    html:        str           = ""
    # If 0, timestamp = now
    ms_since_epoch: int = 0
    # If empty, use the default avatar icon
    avatar_url: Optional[str] = None


    def __post_init__(self) -> None:
        assert bool(self.sender_id), "No sender_id argument passed."
        assert bool(self.room_id),   "No room_id argument passed."

        if self.markdown and not self.html:
            self.html = markdown.to_html(self.markdown)

        elif self.html and not self.markdown:
            self.markdown = markdown.from_html(self.html)

        elif not self.html and not self.markdown:
            raise TypeError("No markdown or html argument passed.")

        self.linkify_in_html()

        self.ms_since_epoch = self.ms_since_epoch or \
                              QDateTime.currentDateTime().toMSecsSinceEpoch()


    @property
    def _html_class(self) -> str:
        return "message own-message" \
               if self.sender_id == self.receiver_id else "message"

    @property
    def html_avatar(self) -> str:
        "HTML to be displayed by a widget for avatar."

        return ""
        # return "<div class='avatar%s'><img src='%s'></div>" % (
            # self._html_class,
          # self.avatar_url or main_window().icons.path("default_avatar_small")
        # )

    @property
    def html_info(self) -> str:
        "HTML to be displayed by a widget for the info line (name, date, ...)."

        date = QDateTime.fromMSecsSinceEpoch(self.ms_since_epoch)\
               .toString(DATE_FORMAT)

        return (
            f"<div class='info {self._html_class}'>"
            f"<span class='name'>{self.sender_display_name}</span>&nbsp;"
            f"<span class='date'>{date}</span>"
            f"</div>"
        )

    @property
    def html_content(self) -> str:
        "HTML to be displayed by a widget for the message's content."

        return f"<div class='content {self._html_class}'>{self.html}</div>"


    def was_created_before(self, ms_since_epoch: int) -> bool:
        """Return if this message was created before `ms_since_epoch`,
        a unix timestamp in microseconds."""

        return self.ms_since_epoch < ms_since_epoch


    @property
    def sender_display_name(self) -> str:
        "Display name for `self.sender_id`. Tries to avoid network requests."

        try:
            room = main_window().accounts[self.sender_id].rooms[self.room_id]
        except KeyError:
            pass
        else:
            name = room.members_displaynames.get(self.sender_id)
            if name:
                return name

        for client in main_window().accounts.values():
            user = client.users.get(self.sender_id)
            if user:
                return user.get_display_name()

        return self.sender_id


    def linkify_in_html(self) -> None:
        "Wrap all `self.html` URLs in <a> tags if they aren't already."

        html = self.html

        def replacer(match) -> str:
            url = [g for g in match.groups() if g is not None][0]

            re_localhost = r"(?:127\.0\.0\.1|localhost)(?::\d+)?"

            href = "%s%s" % (
                ""        if "://" in url else
                "mailto:" if re.match(r"[^@]+@[^@]+", url) else
                "http://" if re.match(re_localhost, url) else
                "https://",
                url
            )

            for already_linkified in a_tags:
                if href in already_linkified:
                    return url

            return f"<a href='{href}'>{url}</a>"

        # List of <a> tags and anything inside them: `<a …> … <a/ …>`
        a_tags = re.findall(r"<\s*?a(?:\s+\S+)?>[^<]*<\s*/?\s*a(?:\s+\S+)?>",
                            html)

        # Match URL/domains in html like `scheme://…`, `….tld` or `….tld:80/…',
        # IPs and mail addresses:
        reu = (
            r"[A-Za-z]+:///?[^\s<]+|"
            r"(?:[\w_-][\w@._-]*[\w_-]\.[A-Za-z]{2,9}|"
            r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|"
            r"localhost)(?::\d+)?(?:/[^\s<]+)?"
        )

        # Allow URL/domain captures between these characters:
        pairs: List[Tuple[str, str]] = [
            # Begin/end of line, whitespace or HTML tags end/begin
            (r"(?<=^)|(?<=\s)|(?<=>)", r"(?:$|\s|<)"),
            ("<", ">"),
            ("&lt;", "&gt;"),
            ("&gt;", "&lt;"),
            (r"\(", r"\)"),
            (r"\[", r"\]"),
            (r"\{", r"\}"),
        ]

        final_regex = r"(?:%s)" % r"|".join(
            (rf"(?<={p[0]})({reu})(?={p[1]})" for p in pairs)
        )
        self.html = re.sub(final_regex, replacer, html)


    def send(self) -> None:
        "Send this message to the room specified by `self.room_id`."

        room = main_window().accounts[self.sender_id].rooms[self.room_id]
        for func in self.local_echo_hooks.values():
            func(copy(self))
        room.send_html(html=self.html, body=self.markdown)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv2.

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QLabel, QVBoxLayout, QWidget

from . import LOG_PATH, PLEASE_REPORT


class Console(QWidget):
    def __init__(self) -> None:
        super().__init__()

        self.vbox = QVBoxLayout(self)
        self.vbox.setContentsMargins(0, 0, 0, 0)
        self.vbox.setSpacing(0)

        from .display import ConsoleDisplay
        self.display = ConsoleDisplay()

        self.about = QLabel(
            f"Unexpected errors occured.<br>"
            f"A log file is written to <em>{LOG_PATH}</em>.<br>"
            f"{PLEASE_REPORT}"
        )
        self.about.setWordWrap(True)
        self.about.setOpenExternalLinks(True)
        self.about.setTextInteractionFlags(Qt.TextBrowserInteraction)

        self.vbox.addWidget(self.display)
        self.vbox.addWidget(self.about)

        self.scroller = self.display.scroller



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import pdb as actual_pdb
import shlex
from typing import Callable, Dict, Optional

import docopt
from PyQt5.QtCore import pyqtRemoveInputHook

from . import REGISTERED_COMMANDS, register, say, utils
from .. import main_window
from ..chat import Chat, RoomNotJoinedError, UserNotLoggedInError

EVAL_PARSING_HOOKS: Dict[str, Callable[[Chat, str, bool], str]] = {}


@register
# pylint: disable=redefined-builtin
def eval(chat: Chat, args: dict) -> None:
    """Usage: /eval TEXTS... [-u ID|--user ID -p LEVEL|--pdb LEVEL]

    Run `TEXTS` as commands. If a text doesn't start with `/`, `/say` is used.

    Options:
      -u ID, --user ID
        Say/execute `TEXTS` with a different logged-in user account.
        Right click on an account in the Accounts / Rooms pane to copy its ID.

        Multiple users can be specified by separating them with a comma.

        `ID` can also be one of the following:
        - `@`: Will be replaced by your current user;
        - `*`: Will be replaced by all your logged-in users;
        - `&`: Same as above, excluding your current user.

        If you pass an username instead of a proper ID
        (e.g. `alice` instead of `@alice:matrix.org`),
        the server is assumed to be the same as your current user.

      -p LEVEL, --pdb LEVEL
        Enter the Python debugger. `LEVEL` determines where to start:
        - `1`: As soon as possible, when `TEXTS` are about to be parsed;
        - `2`: When `TEXTS` are about to be said/executed.

    Examples:
    ```
      /eval Hi!
      /eval "Message 1 with spaces" "Message 2"
      /eval "/help --full"
      /eval "/eval '/badcommand test'" -p 2

      /eval Hi! -u @my_other_account:matrix.org
      /eval Hi! -u @,account2,account3
      /eval Hi! -u *
    ```"""

    assert args["--pdb"] in (None, "1", "2"), "--pdb must be `1` or `2`."
    pdb_lvl = 0 if not args["--pdb"] else int(args["--pdb"])

    if pdb_lvl == 1:
        pyqtRemoveInputHook()
        actual_pdb.set_trace()  # pylint: disable=no-member

    users = [chat.client.user_id]

    if args["--user"]:
        users      = args["--user"].split(",")
        current    = chat.client.user_id
        all_logged = [uid for uid, client in main_window().accounts.items()
                      if chat.room.room_id in client.rooms]

        all_logged_no_current = [u for u in all_logged if u != current]

        users = [current if u == "@" else u for u in users]

        while "*" in users:
            index = users.index("*")
            users = users[:index] + all_logged + users[index + 1:]

        while "&" in users:
            index = users.index("&")
            users = users[:index] + all_logged_no_current + users[index + 1:]

    for user in users:
        for text in args["TEXTS"]:
            eval_f(chat, text=text, user_id=user, pdb_level=pdb_lvl)


def eval_f(chat:      Chat,
           text:      str,
           user_id:   Optional[str] = None,
           pdb_level: int           = 0) -> None:

    if user_id:
        if user_id.startswith("@") and ":" not in user_id:
            user_id = user_id[1:]

        if not user_id.startswith("@"):
            server  = chat.client.user_id.split(":")[-1]
            user_id = f"@{user_id}:{server}"

        try:
            chat = Chat(user_id, chat.room.room_id)
        except (UserNotLoggedInError, RoomNotJoinedError) as err:
            utils.print_err(chat, str(err))
            return

    force_say = False
    if text.startswith("//") or text.startswith(r"\/"):
        force_say = True
        text      = text[1:]

    for hook in EVAL_PARSING_HOOKS.values():
        text = hook(chat, text, force_say)
        if not text:
            return

    if force_say or not text.startswith("/"):
        say.say_f(chat, text)
        return

    try:
        func, *args = shlex.split(text)
    except ValueError as err:
        utils.print_exception(chat, err)
        return

    try:
        parse_func = REGISTERED_COMMANDS[func.lstrip("/")]
    except KeyError:
        utils.print_err(
            chat,
            f"Command not found: `{func}`. \n"
            f"Type `/help` to see available commands.\n"
            f"Prepend `/` or `\\` to ignore command parsing and say this "
            f"message literally, e.g. `/{func}{' …' if args else ''}`."
        )
        return

    try:
        args = docopt.docopt(parse_func.__doc__, help=False, argv=args)
    except docopt.DocoptExit:
        utils.print_err(
            chat,
            f"Invalid command syntax or bad option, "
            f"see `/help {func.lstrip('/')}`."
        )
    except docopt.DocoptLanguageError as err:
        utils.print_exception(chat, err, parse_func)
    else:
        parse_func(chat, args, _pdb_level=pdb_level)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

r''' OUDATED
Registered functions represent commands.
A command will be started in its own thread to not block the client, unless
`@register_command(run_in_thread=False)` is used.

Functions will receive a `harmonyqt.chat.Chat` object as first argument,
representing the chat from which the user typed the command.
Other arguments will be strings corresponding to what the user typed after
the `/function_name `.

Unless the wanted type for an argument *is* `str`, it must be manually
converted to its appropriate types.  Common types:
- Integer number (convert with `int(arg)`)
- Decimal number (`float(arg)`)
- Boolean (use `str_arg_to_bool(arg)` from harmonyqt.commands.utils)
- List of integer, decimal, string or boolean values
  (use `str_arg_to_list` from harmonyqt.commands.utils).

When taking user ID arguments, use `expand_user(user)` from
`harmonyqt.commands.utils` for each ID.
This provides features like expanding a single `@` to the ID of the user that
typed the command.

A function should have a docstring. The first line will be shown in `/help`,
and the full docstring will be shown with `/help <function name>`.
The docstring can include markdown formatting, except images.

This example command:
```python3
    from harmonyqt.chat import Chat
    from harmonyqt.commands.utils import str_arg_to_list

    @register
    def nudge(chat: Chat, users: str, times: str = "1") -> None:
        """Send an alert to users X times."""
        users_list = [expand_user(u) for u in str_arg_to_list(users)]
        times_int  = int(times)
        ...
```

May be used in the following ways from a chat:

    /nudge @alice:matrix.org
    /nudge @alice:matrix.org,@marisa:matrix.org 3
    /nudge @alice:matrix.org,"value,with,comma or space" times=1
    /nudge times="1" users=@alice:matrix.org,"value,with,comma or space"
    ...

Any value containing spaces (and commas (`,`) for list values) must be put
inside quotes, or the bad character must be escaped using a
backslash (`\`) before it.'''

import functools
import shlex
import pdb as actual_pdb
import traceback
from threading import Thread
from typing import Callable, Dict, List, Optional, Union

from PyQt5.QtCore import pyqtRemoveInputHook

from . import utils
from ..chat import Chat

FuncType = Callable[..., None]

REGISTERED_COMMANDS: Dict[str, FuncType] = {}


def register(func: Optional[FuncType] = None, run_in_thread: bool = True):
    # func will be None if called without parentheses.

    def decorator(func: FuncType):
        def executor(*args, _pdb_level: int = 0, **kwargs) -> None:
            try:
                if _pdb_level == 2:
                    pyqtRemoveInputHook()
                    actual_pdb.set_trace()  # pylint: disable=no-member

                func(*args, **kwargs)
            except Exception as err:
                utils.print_exception(chat=args[0], err=err, bad_func=func)

        @functools.wraps(func)
        def wrapper(chat: Chat, *args, _pdb_level: int = 0,  **kwargs) -> None:
            args_ = [chat] + list(args)

            if _pdb_level > 0 or not run_in_thread:
                executor(*args_, _pdb_level=_pdb_level, **kwargs)
            else:
                Thread(target=executor, args=args_, kwargs=kwargs, daemon=True
                      ).start()

        REGISTERED_COMMANDS[func.__name__] = wrapper
        return wrapper

    # Allow decorator to be called with or without parentheses:
    if callable(func):
        return decorator(func)
    return decorator


# pylint: disable=wrong-import-position,redefined-builtin,reimported
# Standard core commands, cannot be disabled
from . import eval, say, help
# Other commands
from . import alias, autorun, nick, pdb, shell, room_set



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import atexit
import os
import sys
import threading
from traceback import format_exception, print_exc, print_exception
from types import TracebackType
from typing import List, Type

from atomicfile import AtomicFile
from PyQt5.QtCore import QDateTime, QObject, pyqtSignal

from ..__about__ import __email__
from ..utils import get_error_file

EMAIL         = __email__
URL_ISSUES    = "https://github.com/mirukan/harmonyqt/issues"
A_STYLE       = "style = 'color: rgb(40, 160, 110);'"
PLEASE_REPORT = (
    f"Please report problems, including this file, on "
    f"<a href='{URL_ISSUES}' {A_STYLE}>GitHub</a> or "
    f"by email to <a href='mailto:{EMAIL}' {A_STYLE}>{EMAIL}</a>."
)
LOG_PATH   = get_error_file()
WRITE_LOCK = threading.Lock()

# pylint: disable=wrong-import-position
from .console import Console

CAUGTH_EXCEPTION_MAY_EXIT: List[str] = []


def except_hook(type_:     Type[BaseException],
                value:     BaseException,
                traceback: TracebackType) -> None:

    if type_ is KeyboardInterrupt:
        sys.exit(130)

    try:
        print_exception(type_, value, traceback)
    except OSError:  # No terminal connected
        pass

    trace_str = "\n".join(format_exception(type_, value, traceback))
    date      = QDateTime.currentDateTime().toString("yyyy-MM-dd HH:mm:ss.z t")

    add_to_log_file(trace_str, date)

    try:
        from .. import main_window
        window = main_window()

        if window.isVisible() and hasattr(window, "error_dock"):
            CAUGTH_EXCEPTION_MAY_EXIT.append(trace_str)
            window.show_error_dock()
            window.error_dock.widget().display.print_error(trace_str, date)
            return
    except Exception:
        print_exc()

    show_box(trace_str)


def add_to_log_file(trace_str: str, date: str) -> None:
    with WRITE_LOCK:
        mode = "a" if os.path.exists(LOG_PATH) else "w"
        with open(LOG_PATH, mode) as out_file:
            out_file.write(f"{date}\n\n{trace_str}\n\n\n")


def show_box(trace_str: str) -> None:
    from . import boxes
    boxes.FatalErrorBox(trace_str).exec_()
    sys.exit(99)


@atexit.register
def error_box_if_console_exited():
    if not CAUGTH_EXCEPTION_MAY_EXIT:
        return

    err = CAUGTH_EXCEPTION_MAY_EXIT[-1]

    try:
        from .. import main_window
        normally_closed = main_window().normal_close
    except Exception:
        normally_closed = False

    if not normally_closed:
        show_box(trace_str=err)


class ExceptHookFromThreadCaller(QObject):
    call = pyqtSignal(tuple)

    def __init__(self):
        """https://stackoverflow.com/a/31622038
        Make `sys.excepthook` work in threads."""
        super().__init__()
        self.call.connect(lambda info: sys.excepthook(*info))
        real_self = self

        init_original = threading.Thread.__init__

        def init(self, *args, **kwargs):

            init_original(self, *args, **kwargs)
            run_original = self.run

            def run_with_except_hook(*args2, **kwargs2):
                try:
                    run_original(*args2, **kwargs2)
                except Exception:
                    real_self.call.emit(sys.exc_info())

            self.run = run_with_except_hook

        threading.Thread.__init__ = init


sys.excepthook = except_hook
_              = ExceptHookFromThreadCaller()



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import json
import time
from multiprocessing.pool import ThreadPool
from threading import Lock
from typing import Dict, Set

from PyQt5.QtCore import QDateTime, QObject, pyqtSignal

from matrix_client.client import MatrixClient

from . import main_window, message


class _SignalObject(QObject):
    # User ID, event
    new_event        = pyqtSignal(str, dict)
    new_unique_event = pyqtSignal(str, dict)

    new_message = pyqtSignal(message.Message)

    # User ID
    new_account  = pyqtSignal(str)
    account_gone = pyqtSignal(str)
    # User ID, room ID, new display name, new avatar URL
    account_change = pyqtSignal(str, str, str, str)

    # User ID, room ID
    new_room    = pyqtSignal(str, str)
    room_rename = pyqtSignal(str, str)
    left_room   = pyqtSignal(str, str)
    # User ID, room ID, invited by user ID, display name, name, canon alias
    new_invite = pyqtSignal(str, str, str, str, str, str)


class EventManager:
    def __init__(self) -> None:
        self._pool: ThreadPool = ThreadPool(1)

        self.start_ms_since_epoch: int = \
            QDateTime.currentDateTime().toMSecsSinceEpoch()

        self.signals = _SignalObject()

        self._added_rooms:   Dict[str, Set[str]] = {}  # {user_id: {room_id}}
        self._got_event_ids: Dict[str, Set[str]] = {}  # {user_id: {event_id}}

        self._lock = Lock()

        main_window().accounts.signals.login.connect(self.add_account)
        main_window().accounts.signals.logout.connect(self.on_account_logout)
        self.signals.new_room.connect(self.add_room_listeners)


    def add_account(self, client: MatrixClient) -> None:
        "Setup event listeners for client. Called from AccountManager.login()."
        user_id = client.user_id

        client.add_listener(lambda ev, u=user_id: self.process_event(u, ev))

        client.add_presence_listener(
            lambda ev: self.on_presence_event(user_id, ev))

        client.add_ephemeral_listener(
            lambda ev: self.on_ephemeral_event(user_id, ev))

        client.add_invite_listener(
            lambda rid, state: self.on_invite_event(user_id, rid, state))

        client.add_leave_listener(
            lambda rid, _: self.on_leave_event(user_id, rid))

        client.start_listener_thread(timeout_ms        = 10_000,
                                     exception_handler = self.on_listen_error)

        self.signals.new_account.emit(client.user_id)


    @staticmethod
    def on_listen_error(err: BaseException) -> None:
        try:
            print(err)
        except OSError:
            pass
        time.sleep(5)


    def add_room_listeners(self, user_id: str, room_id: str) -> None:
        room = main_window().accounts[user_id].rooms[room_id]

        room.add_listener(
            lambda _, ev: self.process_event(user_id, ev))

        room.add_ephemeral_listener(
            lambda _, ev: self.on_ephemeral_event(user_id, ev))

        room.add_state_listener(
            lambda ev: self.on_state_event(user_id, ev))


    def on_account_logout(self, receiver_id: str) -> None:
        self.signals.account_gone.emit(receiver_id)


    def is_old_event(self, event: dict) -> bool:
        return event["origin_server_ts"] < self.start_ms_since_epoch


    def process_event(self, receiver_id: str, event: dict) -> None:
        self.signals.new_event.emit(receiver_id, event)

        ev        = event
        etype     = event["type"]
        room_id   = event["room_id"]
        old       = self.is_old_event(event)

        with self._lock:
            received = self._got_event_ids.setdefault(receiver_id, set())
            if ev["event_id"] in received:
                return
            received.add(ev["event_id"])

            added_for_recv = self._added_rooms.setdefault(receiver_id, set())
            if room_id not in added_for_recv:
                added_for_recv.add(room_id)
                self.signals.new_room.emit(receiver_id, room_id)

        self.signals.new_unique_event.emit(receiver_id, event)

        if etype == "m.room.message":
            self._pool.apply_async(
                func = self.on_new_message,
                args = (receiver_id, event), error_callback=self._on_emit_err,
            )

        if old:
            return

        if etype == "m.room.member" and ev.get("membership") == "join":
            if ev.get("state_key") in main_window().accounts:
                prev = ev.get("unsigned", {}).get("prev_content")
                new  = ev.get("content")

                if prev and new and prev != new:
                    # This won't update automatically in these cases otherwise
                    user = main_window().accounts[ev["state_key"]].user

                    dispname         = new["displayname"] or ev["state_key"]
                    user.displayname = dispname

                    self.signals.account_change.emit(
                        ev["state_key"], room_id,
                        dispname, new["avatar_url"] or ""
                    )

        if etype in ("m.room.name", "m.room.canonical_alias", "m.room.member"):
            self.signals.room_rename.emit(receiver_id, room_id)


    def on_new_message(self, receiver_id: str, event: dict) -> None:
        ev = event

        try:
            if ev["content"]["msgtype"] != "m.text":
                raise RuntimeError
        except Exception:
            print("\nUnsupported msg event:\n", json.dumps(ev, indent=4))
            return

        msg = message.Message(
            sender_id      = ev["sender"],
            receiver_id    = receiver_id,
            room_id        = ev["room_id"],
            markdown       = ev["content"]["body"],
            html           = ev["content"].get("formatted_body", ""),
            ms_since_epoch = ev["origin_server_ts"],
        )
        self.signals.new_message.emit(msg)


    @staticmethod
    def _on_emit_err(err: BaseException) -> None:
        raise err


    def on_presence_event(self, receiver_id: str, event: dict) -> None:
        self._log("yellow", "unhandled presence event", receiver_id, event)


    def on_ephemeral_event(self, receiver_id: str, event: dict) -> None:
        self._log("purple", "unhandled ephemeral event", receiver_id, event)


    def on_state_event(self, receiver_id: str, event: dict) -> None:
        self._log("cyan", "unhandled state event", receiver_id, event)


    def on_invite_event(self, receiver_id: str, room_id: int, state: dict
                       ) -> None:
        invite_by = state["events"][-1]["sender"]

        name = alias = ""
        members = []

        for ev in state["events"]:
            etype = ev["type"]

            if etype == "m.room.name":
                name = ev["content"]["name"]

            if etype == "m.room.canonical_alias":
                alias = ev["content"]["alias"]

            if etype == "m.room.member" and ev["state_key"] != receiver_id:
                members.append(ev["content"]["displayname"] or
                               ev["state_key"])

        dispname = name or alias

        if not dispname:
            if not members:
                dispname = "Empty room"
            elif len(members) == 1:
                dispname = members[0]
            elif len(members) == 2:
                dispname = " and ".join(members)
            else:
                members.sort()
                dispname = f"{members[0]} and {len(members) - 1} others"

        self.signals.new_invite.emit(
            receiver_id, room_id, invite_by, dispname, name, alias
        )


    def on_leave_event(self, receiver_id: str, room_id: str) -> None:
        with self._lock:
            self._added_rooms[receiver_id].discard(room_id)
            self.signals.left_room.emit(receiver_id, room_id)


    def _log(self, color: str, *args, force: bool = False) -> None:
        if not force:
            return

        jsons = [json.dumps(arg, indent=4, sort_keys=True) for arg in args]
        nums  = {"black": 0, "red": 1, "green": 2, "yellow": 3, "blue": 4,
                 "purple": 5, "magenta": 5, "cyan": 6, "white": 7, "gray": 7}

        with self._lock:
            print(f"\033[3{nums[color]}m", *jsons, "\033[0m",
                  sep="\n", end="\n\n")



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import threading

from . import register, utils
from ..chat import Chat
from ..message import Message


@register
def say(chat: Chat, args: dict) -> None:
    """
    Usage:
      /devices [USER_ID]
               [DEVICES_ID...
                [-v|--verify | -b|--blacklist | -i|--ignore | --delete]]

    List, (un)trust or delete user devices.

    If `USER_ID` is unspecified, your current user is used.
    If no `DEVICES_ID` is specified, your/`USER_ID`'s devices will be listed.
    Options will apply to the specified `DEVICES_ID`.

    To verify that one of your device can be trusted:
      - From that device, check its name, ID and key by e.g.
        typing `/devices` from Harmony or going to User Settings in Riot
      - Ensure that these match with what you see by typing `/devices` here.

    To verify that someone else's device can be trusted:
      - Ask the owner for the device's name, ID and key they see
      - Ensure that these match with what you see by typing `/devices` here.

    Devices that you don't recognize or that fails the check above
    could be attackers eavesdropping and should be blacklisted.

    Options:
      -t, --trust
        Mark `DEVICES_ID` as verified.
        In encrypted rooms, messages are by default only sent to your and
        other users's verified devices.

      -b, --blacklist
        Mark `DEVICES_ID` as blacklisted.
        In encrypted rooms, messages will never be sent to these devices.

      -i, --ignore
        Allow sending encrypted messages to `DEVICES_ID` regardless of
        if they are untrusted.

    Examples:
    ```
    ```"""

    for msg in args["MESSAGES"]:
        say_f(chat=chat, text=msg, is_html=args["--html"], echo=args["--echo"])


def say_f(chat: Chat, text: str, is_html: bool = False, echo: bool = False
         ) -> None:
    if not text.strip():
        return

    if echo:
        utils.print_info(chat, text, is_html=is_html)
        return


    msg = Message(
        sender_id = chat.client.user_id,
        room_id   = chat.room.room_id,
        markdown  = text if not is_html else "",
        html      = text if is_html     else "",
    )

    # Prevent mixing up if user sends msg B before msg A is received
    # by the server:
    with SEND_LOCK:
        msg.send()



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import traceback

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import (
    QFontMetrics, QResizeEvent, QTextCursor, QTextTableFormat
)
from PyQt5.QtWidgets import QTextBrowser

from harmonyqt.scroller import Scroller


class MessageDisplay(QTextBrowser):
    # Useful for anything in a thread that wants to show text
    system_print_request = pyqtSignal(str, str, bool)  # text, level, is_html


    def __init__(self) -> None:
        super().__init__()

        self.scroller = Scroller(self)
        self._set_previous_resize_vbar()
        self.scroller.vbar.valueChanged.connect(
            lambda _: self._set_previous_resize_vbar()
        )

        doc = self.document()
        doc.setUndoRedoEnabled(False)

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        # self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setOpenExternalLinks(True)

        self.font_height         = QFontMetrics(doc.defaultFont()).height()
        self.system_print_format = QTextTableFormat()
        self.system_print_format.setBorder(0)
        self.system_print_format.setTopMargin(self.font_height)

        self.system_print_request.connect(self.system_print)


    def apply_style(self) -> None:
        try:
            from harmonyqt import main_window
            window = main_window()
            doc    = self.document()
            doc.setDefaultStyleSheet(window.theme.style("messages"))
        except Exception:
            traceback.print_exc()


    def _set_previous_resize_vbar(self) -> None:
        sc                         = self.scroller
        self._previous_resize_vbar = (sc.vmin, sc.v, sc.vmax)


    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)
        if event.oldSize().height() < 0:
            return

        _, old_vbar_val, old_vbar_max = self._previous_resize_vbar
        scr = self.scroller
        scr.vset(scr.vmax - (old_vbar_max - old_vbar_val))


    def system_print(self,
                     text:    str,
                     level:   str  = "info",
                     is_html: bool = False) -> None:
        assert level in ("info", "warning", "error")
        distance_from_bottom = self.scroller.vmax - self.scroller.v

        html = text

        if not is_html:
            try:
                from harmonyqt import markdown
                html = markdown.to_html(text)
            except Exception:
                traceback.print_exc()

        html = f"<div class='system {level}'>{html}</div>"

        cursor = QTextCursor(self.document())
        cursor.beginEditBlock()
        cursor.movePosition(QTextCursor.End)
        cursor.insertTable(1, 1, self.system_print_format)
        cursor.insertHtml(html)
        cursor.endEditBlock()

        if distance_from_bottom <= 10:
            self.scroller.go_min_left().go_bottom()



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from typing import Any, Callable, Dict, List, Tuple

from harmonyqt import app, main, shortcuts

from .scroller import Scroller

SHORTCUTS: Dict[str, Tuple[Callable[[Scroller], Any], List[str]]] = {
    # Alt+direction simple scroll
    "left":  (lambda s: s.go_left(),  ["A-h", "A-Left"]),
    "down":  (lambda s: s.go_down(),  ["A-j", "A-Down"]),
    "up":    (lambda s: s.go_up(),    ["A-k", "A-Up"]),
    "right": (lambda s: s.go_right(), ["A-l", "A-Right"]),

    # Ctrl+Alt+direction page scroll
    "page left":  (lambda s: s.go_page_left(),  ["C-A-h", "C-A-Left"]),
    "page down":  (lambda s: s.go_page_down(),  ["C-A-j", "C-A-Down"]),
    "page up":    (lambda s: s.go_page_up(),    ["C-A-k", "C-A-Up"]),
    "page right": (lambda s: s.go_page_right(), ["C-A-l", "C-A-Right"]),

    # Ctrl+Alt+Shift+direction begin/end scroll
    "to left edge":  (lambda s: s.go_min_left(),  ["C-A-S-h", "C-A-S-Left"]),
    "to bottom":     (lambda s: s.go_bottom(),    ["C-A-S-j", "C-A-S-Down"]),
    "to top":        (lambda s: s.go_top(),       ["C-A-S-k", "C-A-S-Up"]),
    "to right edge": (lambda s: s.go_max_right(), ["C-A-S-l", "C-A-S-Right"]),
}


def register(window) -> None:
    for name, (func, binds) in SHORTCUTS.items():

        def on_activation(func = func) -> None:
            widget = app().get_focused(lambda w: hasattr(w, "scroller"))
            if widget is not None:
                func(widget.scroller)

        window.shortcuts.add(shortcuts.Shortcut(
            name          = f"Scroll {name}",
            on_activation = on_activation,
            bindings      = binds,
        ))


main.HOOKS_INIT_2_BEFORE_LOGIN["Register Scroller shortcuts"] = register



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from PyQt5.QtWidgets import QWidget


class Scroller:
    def __init__(self, widget: QWidget) -> None:
        self.widget = widget
        self.hbar   = widget.horizontalScrollBar()
        self.vbar   = widget.verticalScrollBar()


    @property
    def hmin(self) -> int:
        return self.hbar.minimum()

    @property
    def vmin(self) -> int:
        return self.vbar.minimum()

    @property
    def h(self) -> int:
        return self.hbar.value()

    @property
    def v(self) -> int:
        return self.vbar.value()

    @property
    def hmax(self) -> int:
        return self.hbar.maximum()

    @property
    def vmax(self) -> int:
        return self.vbar.maximum()


    @property
    def hstep(self) -> int:
        return self.hbar.singleStep()

    @property
    def vstep(self) -> int:
        return self.vbar.singleStep()

    @property
    def hstep_page(self) -> int:
        return self.hbar.pageStep()

    @property
    def vstep_page(self) -> int:
        return self.vbar.pageStep()


    def hset(self, to: int) -> "Scroller":
        self.hbar.setValue(to)
        return self

    def vset(self, to: int) -> "Scroller":
        self.vbar.setValue(to)
        return self


    def go_left(self, times: int = 1) -> "Scroller":
        return self.hset(self.h - self.hstep * times)

    def go_right(self, times: int = 1) -> "Scroller":
        return self.hset(self.h + self.hstep * times)

    def go_up(self, times: int = 1) -> "Scroller":
        return self.vset(self.v - self.vstep * times)

    def go_down(self, times: int = 1) -> "Scroller":
        return self.vset(self.v + self.vstep * times)


    def go_page_left(self, times: int = 1) -> "Scroller":
        return self.hset(self.h - self.hstep_page * times)

    def go_page_right(self, times: int = 1) -> "Scroller":
        return self.hset(self.h + self.hstep_page * times)

    def go_page_up(self, times: int = 1) -> "Scroller":
        return self.vset(self.v - self.vstep_page * times)

    def go_page_down(self, times: int = 1) -> "Scroller":
        return self.vset(self.v + self.vstep_page * times)


    def go_min_left(self) -> "Scroller":
        return self.hset(self.hmin)

    def go_max_right(self) -> "Scroller":
        return self.hset(self.hmax)

    def go_top(self) -> "Scroller":
        return self.vset(self.vmin)

    def go_bottom(self) -> "Scroller":
        return self.vset(self.vmax)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv2.

from multiprocessing.pool import ThreadPool

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QWidget

from . import Chat
from .. import dock, main_window


class ChatDock(dock.Dock):
    def __init__(self, user_id: str, room_id: str, parent: QWidget) -> None:
        self.user_id: str = user_id
        self.room_id: str = room_id
        super().__init__(f"{self.user_id}: {self.room_id}",
                         parent=parent, can_hide_title_bar=False)

        self._pool = ThreadPool(1)
        self.autoset_title()

        self.change_room(self.user_id, self.room_id)

        self.visibilityChanged.connect(self.on_visibility_change)
        self.dockLocationChanged.connect(self.on_location_change)


    def autoset_title(self) -> None:
        def get() -> str:
            client = main_window().accounts[self.user_id]
            return ": ".join((client.user.get_display_name(),
                              client.rooms[self.room_id].display_name))

        def set_(title: str) -> None:
            self.title = title

        self._pool.apply_async(get, callback=set_)


    def change_room(self, to_user_id: str, to_room_id: str) -> None:
        main_window().visible_chat_docks.pop((self.user_id, self.room_id),
                                             None)

        self.user_id, self.room_id = to_user_id, to_room_id
        self.chat = Chat(self.user_id, self.room_id)
        self.setWidget(self.chat)
        self.autoset_title()

        main_window().visible_chat_docks[(self.user_id, self.room_id)] = self


    def focus(self) -> None:
        super().focus()
        self.chat.send_area.box.setFocus()


    def on_visibility_change(self, visible: bool) -> None:
        ids = (self.user_id, self.room_id)
        if visible:
            main_window().visible_chat_docks[ids] = self
        else:
            main_window().visible_chat_docks.pop(ids, None)


    def on_location_change(self, _: Qt.AllDockWidgetAreas) -> None:
        # Needed for situations where user drags a dock then opens a new
        # chat, since the location of that new chat dock is dependent on
        # the last focused dock.
        self.focus()



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import traceback
from typing import Callable, Optional

from ..chat import Chat


def print_info(chat: Chat, text: str, is_html: bool = False) -> None:
    chat.messages.system_print_request.emit(text, "info", is_html)


def print_warn(chat: Chat, text: str, is_html: bool = False) -> None:
    chat.messages.system_print_request.emit(text, "warning", is_html)


def print_err(chat: Chat, text: str, is_html: bool = False) -> None:
    chat.messages.system_print_request.emit(text, "error", is_html)


def print_exception(chat:     Chat,
                    err:      Exception,
                    bad_func: Optional[Callable] = None) -> None:
    traceback.print_exc()
    trace = traceback.format_exc().rstrip()
    shown = "".join((
        f"<code>{bad_func.__name__}</code>: " if bad_func else "",
        f"{type(err).__name__} - {err!s}<br><pre><code>{trace}</pre></code>",
    ))
    print_err(chat, shown, is_html=True)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import pdb as actual_pdb

from PyQt5.QtCore import pyqtRemoveInputHook

from . import register
from .. import main_window, app
from ..chat import Chat


@register(run_in_thread=False)
def pdb(ch: Chat, _) -> None:
    """Usage: /pdb

    Enter the Python debugger. Harmony must be running from a terminal."""

    ap = app()
    wi = main_window()
    print("[App]        ap =", ap)
    print("[MainWindow] wi =", wi)
    print("[Chat]       ch =", ch)

    pyqtRemoveInputHook()
    actual_pdb.set_trace()  # pylint: disable=no-member



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from multiprocessing.pool import ThreadPool
from typing import Dict, List

from PyQt5.QtCore import QObject, QPoint, Qt, pyqtSignal
from PyQt5.QtGui import QIcon, QKeyEvent, QMouseEvent
from PyQt5.QtWidgets import (
    QAction, QHeaderView, QSizePolicy, QTreeWidget, QTreeWidgetItem
)

from harmonyqt import __about__, actions, app, main_window, shortcuts
from harmonyqt.dialogs import AcceptRoomInvite
from harmonyqt.menu import Menu
from harmonyqt.scroller import Scroller
from matrix_client.client import MatrixClient
from matrix_client.room import Room


class UserTree(QTreeWidget):
    def __init__(self) -> None:
        super().__init__()
        self.accounts:   Dict[str, "AccountRow"] = {}
        self.blank_rows: List["BlankRow"]        = []

        self.scroller: Scroller = Scroller(self)

        main_window().shortcuts.add(shortcuts.Shortcut(
            name          = "Focus Accounts / Rooms",
            on_activation = self.setFocus,
            bindings      = ["A-a"],
            autorepeat    = False
        ))

        self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Expanding)
        self.setColumnCount(2)  # avatar/name; unread msg num/invite indicator
        self.setUniformRowHeights(True)
        self.setAnimated(True)
        self.setAutoExpandDelay(500)
        self.setHeaderHidden(True)  # TODO: customizable cols
        self.setSelectionMode(QTreeWidget.ExtendedSelection)
        self.setExpandsOnDoubleClick(False)  # Handled by signals/events
        self.setIndentation(12)

        self.header().setMinimumSectionSize(1)
        self.header().setStretchLastSection(False)
        self.header().setSectionResizeMode(0, QHeaderView.Stretch)
        self.header().setSectionResizeMode(1, QHeaderView.ResizeToContents)

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.on_context_menu_request)

        ev_sig = main_window().events.signals
        ev_sig.new_account.connect(self.add_account)
        ev_sig.account_gone.connect(self.del_account)
        ev_sig.new_room.connect(self.on_add_room)
        ev_sig.new_invite.connect(self.on_add_room)
        ev_sig.room_rename.connect(self.on_rename_room)
        ev_sig.left_room.connect(self.on_left_room)
        ev_sig.account_change.connect(self.on_account_change)

        self.actions: List[QAction] = []
        self.init_actions()


    def init_actions(self) -> None:
        win = main_window()
        self.actions = [
            actions.AddAccount(win),
            actions.NewChat(win),
            actions.SetStatus(win),
        ]


    def get_context_menu_actions(self) -> List[QAction]:
        return self.actions


    def add_account(self, user_id: str) -> None:
        if user_id in self.accounts:
            return

        row = AccountRow(self, user_id)
        row.signals.ui_updated.connect(self.sort_account_rows)
        self.accounts[user_id] = row


    def sort_account_rows(self) -> None:
        root = self.invisibleRootItem()
        root.sortChildren(0, Qt.AscendingOrder)

        for row in self.blank_rows:
            root.removeChild(row)

        for row in self.accounts.values():
            index = root.indexOfChild(row)
            if index > 0:
                blank = BlankRow()
                root.insertChild(index, blank)
                self.blank_rows.append(blank)


    def del_account(self, user_id: str) -> None:
        if user_id not in self.accounts:
            return

        to_del: AccountRow = self.accounts[user_id]
        root               = self.invisibleRootItem()

        for row in self.blank_rows:
            if root.indexOfChild(row) == root.indexOfChild(to_del) - 1:
                root.removeChild(row)

        root.removeChild(to_del)
        del self.accounts[user_id]

        first = root.child(0)
        if isinstance(first, BlankRow):
            root.removeChild(first)


    def on_context_menu_request(self, position: QPoint) -> None:
        selected = [self.itemFromIndex(s) for s in self.selectedIndexes()
                    if s.column() == 0 and not isinstance(s, BlankRow)]

        acts: List[QAction] = []
        for row in selected:
            if hasattr(row, "get_context_menu_actions"):
                acts += row.get_context_menu_actions()

        if not selected:
            acts += self.get_context_menu_actions()

        menu = Menu(self, acts)
        menu.exec_if_not_empty(self.mapToGlobal(position))


    def on_add_room(self, user_id: str, room_id: str,
                    invite_by: str = "", display_name: str = "",
                    name:      str = "", alias:        str = "") -> None:
        self.accounts[user_id].add_room(room_id,
                                        invite_by, display_name, name, alias)


    def on_rename_room(self, user_id: str, room_id: str) -> None:
        rooms = self.accounts[user_id].rooms
        if room_id in rooms:
            rooms[room_id].update_ui()


    def on_left_room(self, user_id: str, room_id: str) -> None:
        self.accounts[user_id].del_room(room_id)


    def on_account_change(self, user_id: str, _: str,
                          new_display_name: str, new_avatar_url: str) -> None:
        self.accounts[user_id].update_ui(new_display_name, new_avatar_url)


    def really_clear_selection(self) -> None:
        self.clearSelection()
        self.setCurrentItem(None)


    def keyPressEvent(self, event: QKeyEvent) -> None:
        super().keyPressEvent(event)

        if event.key() in (Qt.Key_Return, Qt.Key_Enter):  # Enter = keypad
            self.activate_row(self.currentItem(), event.modifiers())

        elif event.key() == Qt.Key_Escape:
            self.really_clear_selection()


    def mousePressEvent(self, event: QMouseEvent) -> None:
        super().mousePressEvent(event)

        btn  = event.button()
        mods = app().keyboardModifiers()

        if btn == Qt.RightButton or \
           btn == Qt.LeftButton and mods != Qt.NoModifier:
            return

        clicked_row = self.itemAt(event.pos())
        self.activate_row(clicked_row, mods, btn == Qt.MiddleButton)

        if not clicked_row:
            self.really_clear_selection()


    def activate_row(self,
                     row:             QTreeWidgetItem,
                     kb_modifiers:    Qt.KeyboardModifiers,
                     middle_click: bool = False) -> None:
        if hasattr(row, "on_activation"):
            row.on_activation(kb_modifiers, middle_click)
            self.really_clear_selection()


class BlankRow(QTreeWidgetItem):
    def __init__(self) -> None:
        # Don't take a parent or sorting won't work correctly
        super().__init__()
        self.setFlags(Qt.NoItemFlags | Qt.ItemNeverHasChildren)


class _AccountRowSignals(QObject):
    ui_updated = pyqtSignal()


class AccountRow(QTreeWidgetItem):
    def __init__(self, parent: UserTree, user_id: str) -> None:
        super().__init__(parent)
        self.user_tree: UserTree           = parent
        self.client:    MatrixClient       = main_window().accounts[user_id]
        self.rooms:     Dict[str, RoomRow] = {}

        self.signals = _AccountRowSignals()
        self._pool   = ThreadPool(1)

        self.auto_expanded_once: bool = False
        self.update_ui()

        self.actions: List[QAction] = []
        self.init_actions()


    def init_actions(self) -> None:
        self.actions = [
            actions.NewChat(self.user_tree, self.client.user_id),
            actions.DelAccount(self.user_tree, self.client.user_id),
        ]


    def get_context_menu_actions(self) -> List[QAction]:
        return self.actions


    def update_ui(self, new_display_name: str = "", _: str = "") -> None:
        self.setToolTip(0, self.client.user_id)

        def update(name: str) -> None:
            self.setText(0, name)
            self.signals.ui_updated.emit()

        if new_display_name:
            update(new_display_name)
        else:
            self._pool.apply_async(self.client.user.get_display_name,
                                   callback = update)


    def on_activation(self, *_) -> None:
        expanded = self.isExpanded()
        self.setExpanded(not expanded)


    def add_room(self, room_id: str,
                 invite_by: str = "", display_name: str = "",
                 name:      str = "", alias:        str = "") -> None:
        if room_id in self.rooms:
            self.rooms[room_id].update_ui()
            return

        self.rooms[room_id] = RoomRow(self, room_id,
                                      invite_by, display_name, name, alias)

        if not self.auto_expanded_once:
            self.setExpanded(True)  # TODO: unless user collapsed manually
            self.auto_expanded_once = True


    def del_room(self, room_id: str) -> None:
        if room_id in self.rooms:
            self.removeChild(self.rooms[room_id])
            del self.rooms[room_id]


    def sort_room_rows(self) -> None:
        self.sortChildren(0, Qt.AscendingOrder)


class _RoomRowSignals(QObject):
    ui_updated = pyqtSignal()


class RoomRow(QTreeWidgetItem):
    def __init__(self, parent: AccountRow, room_id: str,
                 invite_by: str = "", display_name: str = "",
                 name:      str = "", alias:        str = "") -> None:
        super().__init__(parent)
        self.setFlags(self.flags() | Qt.ItemNeverHasChildren)

        self.account_row: AccountRow = parent
        self.invite_by:   str        = invite_by

        self.signals = _RoomRowSignals()
        self.signals.ui_updated.connect(self.account_row.sort_room_rows)

        self._pool = ThreadPool(1)

        if invite_by:
            self.room: Room           = Room(parent.client, room_id)
            self.room.name            = self.room.name            or name
            self.room.canonical_alias = self.room.canonical_alias or alias
        else:
            self.room: Room = parent.client.rooms[room_id]

        self.setTextAlignment(1, Qt.AlignRight)  # msg unread/invite indicator
        self.update_ui(display_name)

        self.actions_invite: List[QAction] = []
        self.actions_normal: List[QAction] = []
        self.init_actions()


    def init_actions(self) -> None:
        tree    = self.account_row.user_tree
        user_id = self.account_row.client.user_id

        self.actions_invite = [
            actions.AcceptInvite(tree, self.room, self.on_invite_accept),
            actions.DeclineInvite(tree, self.room, self.on_leave),
        ]
        self.actions_normal = [
            actions.InviteToRoom(tree, self.room, user_id),
            actions.LeaveRoom(tree, self.room, self.on_leave),
        ]


    def get_context_menu_actions(self) -> List[QAction]:
        return self.actions_invite if self.invite_by else self.actions_normal


    def update_ui(self, invite_display_name: str = "") -> None:
        tooltips = self.room.aliases + [self.room.room_id]

        if self.invite_by:
            self.setIcon(1, main_window().icons.icon("indicator_invite"))
            tooltips.insert(0, f"Pending invitation from {self.invite_by}")
        else:
            self.setIcon(1, QIcon())

        tooltips.append(
            "\nMiddle click to open in a new tab\n"
            "Shift + Middle click to open in a horizontal split\n"
            "Ctrl + Middle click to open in a vertical split"
        )

        tooltips = "\n".join(tooltips)
        for col in range(self.columnCount()):
            self.setToolTip(col, tooltips)

        def update(name: str) -> None:
            self.setText(0, name)
            self.signals.ui_updated.emit()

        if invite_display_name:
            update(invite_display_name)
        else:
            if not self.text(0).strip():
                self.setText(0, self.room.room_id)

            # Raises exception for rooms we're invited to but not joined
            self._pool.apply_async(lambda: self.room.display_name,
                                   callback = update)


    def on_activation(self,
                      kb_modifiers:    Qt.KeyboardModifiers,
                      middle_click: bool = False) -> None:
        user_id = self.account_row.client.user_id
        room_id = self.room.room_id

        if self.invite_by:
            dialog = AcceptRoomInvite(
                main_window(), self.room, self.text(0), self.invite_by,
            )
            dialog.exec()

            if dialog.clickedButton() is dialog.yes:
                self.on_invite_accept()
            elif dialog.clickedButton() is dialog.no:
                self.on_leave()
                return
            else:
                return

        in_new = ""
        orient = Qt.Horizontal
        if middle_click and kb_modifiers == Qt.ShiftModifier:
            in_new = "split"
        elif middle_click and kb_modifiers == Qt.ControlModifier:
            in_new = "split"
            orient = Qt.Vertical
        elif middle_click:
            in_new = "tab"

        main_window().go_to_chat_dock(user_id, room_id,
                                      in_new=in_new, split_orientation=orient)


    def on_invite_accept(self) -> None:
        self.invite_by = ""
        self.update_ui()


    def on_leave(self) -> None:
        self.account_row.del_room(self.room.room_id)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from collections import OrderedDict
from typing import Dict, List, Sequence, Set, Tuple

from PyQt5.QtCore import QPoint
from PyQt5.QtWidgets import QAction, QMenu, QWidget


class Menu(QMenu):
    def __init__(self, parent: QWidget, actions: Sequence[QAction] = ()
                ) -> None:
        super().__init__(parent)
        self.is_empty: bool = True

        self.setWindowOpacity(0.9)
        # self.setTearOffEnabled(True)
        self.setToolTipsVisible(True)

        if not actions:
            return

        # Auto-group "duplicate" actions
        # e.g. if two room rows are selected, show only one menu entry:
        # "Leave selected rooms"

        from .actions import Action, MultiselectAction

        seen_once: Dict[str, Action]       = {}
        groups:    Dict[str, List[Action]] = {}
        to_add:    List[Action]            = []

        for act in actions:
            ms_txt = getattr(act, "multiselect_text", None)

            if not ms_txt:
                continue

            if ms_txt not in seen_once:
                seen_once[ms_txt] = act
                continue

            if ms_txt not in groups:
                groups[ms_txt] = [seen_once[ms_txt]]

            groups[ms_txt].append(act)

        for acts in groups.values():
            to_add.append(MultiselectAction(acts))

        for act in actions:
            ms_txt = getattr(act, "multiselect_text", None)

            if not ms_txt or ms_txt not in groups:
                to_add.append(act)


        # Remove left duplicates: action without multiselect_text
        seen_once2: Set[Tuple[str, str]]          = set()
        to_add2:    Dict[Tuple[str, str], Action] = OrderedDict()

        for act in to_add:
            ident: Tuple[str, str] = (type(act).__name__, act.text)

            if ident in seen_once:
                to_add2.pop(ident, None)
                continue

            to_add2[ident] = act
            seen_once2.add(ident)

        if to_add2:
            self.is_empty = False
            self.addActions(to_add2.values())


    def exec_if_not_empty(self, position: QPoint) -> None:
        if not self.is_empty:
            self.exec_(position)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import traceback

from PyQt5.QtWidgets import QMessageBox, QTextEdit

from . import PLEASE_REPORT, LOG_PATH


class ErrorBox(QMessageBox):
    def __init__(self, trace_str: str) -> None:
        super().__init__()
        try:
            from .. import main_window
            window = main_window()
            if hasattr(window, "theme"):
                self.setStyleSheet(window.theme.style("interface"))
                self.setWindowOpacity(0.9)
        except Exception:
            traceback.print_exc()

        self.setWindowTitle("Harmony - Unexpected error")
        self.setText("An unexpected error occured.")
        self.setInformativeText(f"A log file has been written to:<br>"
                                f"<em>{LOG_PATH}</em>.<br><br>"
                                f"{PLEASE_REPORT}")
        self.setDetailedText(trace_str)
        self.setIcon(QMessageBox.Warning)
        self.setStandardButtons(QMessageBox.Close)

        for details in self.findChildren(QTextEdit):
            details.setFixedSize(640, 360)

        for button in self.buttons():
            if self.buttonRole(button) == QMessageBox.ActionRole:
                button.click()
                break


class FatalErrorBox(ErrorBox):
    def __init__(self, trace_str: str) -> None:
        super().__init__(trace_str)
        self.setWindowTitle("Harmony - Fatal error")
        self.setText("Fatal error, execution halted")
        self.setIcon(QMessageBox.Critical)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import json
from typing import Any, Callable, Dict

from matrix_client.client import MatrixClient
from matrix_client.errors import MatrixRequestError
from matrix_client.room import Room

from . import register, utils
from ..chat import Chat

SETTINGS_FUNC: Dict[str, Callable[[MatrixClient, Room, Any], bool]] = {
    "name":          lambda c, r, v: r.set_room_name(v),
    "topic":         lambda c, r, v: r.set_room_topic(v),
    "alias_add":     lambda c, r, v: c.api.set_room_alias(r.room_id, v),
    "alias_del":     lambda c, r, v: c.api.remove_room_alias(v),
    "allow_guests":  lambda c, r, v: r.set_guest_access(True),
    "refuse_guests": lambda c, r, v: r.set_guest_access(False),
    "public":        lambda c, r, v: r.set_invite_only(False),
    "private":       lambda c, r, v: r.set_invite_only(True),
    "enable_e2e":    lambda c, r, v: r.enable_encryption(),
}


@register
def room_set(chat: Chat, args: dict) -> None:
    """Usage: /room_set [options]

    Change settings for the current room (see `/help room_set`).

    Options:
      -n NAME, --name NAME
        Set this room's display name.

      -t TOPIC, --topic TOPIC
        Set this room's description.

      -a ALIAS, --alias-add ALIAS
        Add an #alias for this room.
        If the room is public, users can join it easily using an alias.

      -A ALIAS, --alias-del ALIAS
        Remove a previously set #alias for this room.

      -g, --allow-guests
        Allow guests (unregistered accounts) to join this room.

      -G, --refuse-guests
        Prevent guests from joining this room.

      -p, --public
        Anyone can join this room if they know its ID, link or alias.

      -P, --private
        Require an invitation to join this room.

      --enable-e2e
        Enable end-to-end encryption for this room. Be warned:
        - Once enabled in a room, it can never be disabled again
        - Encryption is still in development
        - New users/devices will not be able to read the room's history from
          before they joined
        - Encrypted messages will only be visible to clients that supports it,
          e.g. Riot and Harmony.

    Examples:
    ```
      /room_set -n "Test room" -t "Just a test"
      /room_set --refuse-guests --private --enable-e2e
    ```"""

    for arg, val in args.items():
        if arg[2:].replace("-", "_") not in SETTINGS_FUNC:
            utils.print_err(chat, f"Unknown setting: {arg!r}.")
            return

    one_passed = False

    for arg, val in args.items():
        if not (val is True or isinstance(val, str)):
            continue

        one_passed = True

        func_name = arg[2:].replace("-", "_")

        try:
            result = SETTINGS_FUNC[func_name](chat.client, chat.room, val)
        except MatrixRequestError as err:
            data = json.loads(err.content)
            utils.print_err(chat, data["error"].replace("don't", "do not"))
        else:
            if result is False:
                utils.print_err(
                    chat,
                    f"You do not have permission to set `{arg}` in this room"
                )

    if not one_passed:
        utils.print_warn(chat, "No option passed, see `/help room_set`.")



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from collections import Mapping
from typing import Callable, List, Optional, Set, Dict

from dataclasses import dataclass, field
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QShortcut, QWidget

from . import main_window

_KEEP_ALIVE: Set[QShortcut]  = set()


@dataclass(frozen=True)
class Shortcut:
    name:          str                = ""
    on_activation: Callable[[], None] = lambda: None
    bindings:      List[str]          = field(default_factory = list)
    autorepeat:    bool               = True
    parent:        Optional[QWidget]  = None
    context:       Qt.ShortcutContext = Qt.WindowShortcut


    def __post_init__(self) -> None:
        for bind in self.bindings:  # pylint: disable=not-an-iterable
            bind = bind.replace("C-", "Ctrl+")\
                        .replace("A-", "Alt+")\
                        .replace("S-", "Shift+")

            qs = QShortcut(bind, self.parent or main_window())
            qs.activated.connect(self.on_activation)  # type: ignore
            qs.setAutoRepeat(self.autorepeat)
            qs.setContext(self.context)
            _KEEP_ALIVE.add(qs)


class ShortcutManager(Mapping):
    def __init__(self) -> None:
        self._registered: Dict[str, Shortcut] = {}

    def __repr__(self) -> str:
        return "%s(%s)" % (type(self).__name__, repr(self._registered))

    def __getitem__(self, key: str) -> Shortcut:
        return self._registered[key]

    def __iter__(self):
        return iter(self._registered)

    def __len__(self) -> int:
        return len(self._registered)


    def add(self, shortcut: Shortcut) -> None:
        self._registered[shortcut.name] = shortcut


    def remove(self, shortcut_name: str) -> None:
        del self._registered[shortcut_name]



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import json
import webbrowser
from typing import Callable, Dict, Optional, Sequence

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QKeySequence
from PyQt5.QtWidgets import QAction, QWidget

from matrix_client.errors import MatrixRequestError
from matrix_client.room import Room

from . import __about__, dialogs, main_window, menu

ImmediateFunc = Optional[Callable[[], None]]

KEYS_BOUND: Dict[str, str] = {}


class KeysAlreadyBoundError(Exception):
    def __init__(self, key: str, action: str, bound_to: str) -> None:
        super().__init__(f"Can't set {key!r} for action {action!r}, "
                         f"already bound to {bound_to!r}.")
        self.key      = key
        self.action   = action
        self.bound_to = bound_to


# General functions

def safe_bind(action: QAction, key: str) -> str:
    name = type(action).__name__

    if key in KEYS_BOUND and KEYS_BOUND[key] != name:
        raise KeysAlreadyBoundError(key, name, KEYS_BOUND[key])

    KEYS_BOUND[key] = name
    return key


# Base classes

class Action(QAction):
    def __init__(self,
                 parent:              QWidget,
                 text:                str           = "",
                 tooltip:             str           = "",
                 icon:                str           = "",
                 shortcut:            str           = "",
                 multiselect_text:    str           = "",
                 multiselect_tooltip: str           = "",
                 immediate_func:      ImmediateFunc = None) -> None:
        super().__init__(parent)
        self.parent              = parent
        self.text                = text
        self.tooltip             = "\n".join((tooltip, shortcut)).strip()
        self.icon_str            = icon
        self.shortcut_str        = shortcut
        self.multiselect_text    = multiselect_text
        self.multiselect_tooltip = multiselect_tooltip
        self.immediate_func      = immediate_func

        self.setText(self.text)

        if self.tooltip:
            self.setToolTip(tooltip)

        if icon:
            self.setIcon(main_window().icons.icon(icon))

        if shortcut:
            self.setShortcutContext(Qt.ApplicationShortcut)
            self.setShortcut(QKeySequence(safe_bind(self, shortcut)))

            try:  # Qt >= 5.10
                self.setShortcutVisibleInContextMenu(True)
            except AttributeError:
                pass

        self.triggered.connect(self._on_trig)

    def _on_trig(self, checked: bool) -> None:
        if self.immediate_func:
            self.immediate_func()

        self.on_trigger(checked)

    def on_trigger(self, checked: bool) -> None:
        pass


class MultiselectAction(Action):
    def __init__(self, actions: Sequence[Action]) -> None:
        super().__init__(
            parent   = actions[0].parent,
            text     = actions[0].multiselect_text,
            tooltip  = actions[0].multiselect_tooltip or actions[0].tooltip,
            icon     = actions[0].icon_str,
            shortcut = actions[0].shortcut_str,
        )
        self.actions         = actions
        self.immediate_funcs = [a.immediate_func for a in actions
                                if a.immediate_func]

    def _on_trig(self, checked: bool) -> None:
        for func in self.immediate_funcs:
            func()
        self.on_trigger(checked)

    def on_trigger(self, checked: bool) -> None:
        for act in self.actions:
            act.on_trigger(checked)

# Rooms

class NewChat(Action):
    def __init__(self, parent: QWidget, for_user_id: str = "") -> None:
        super().__init__(
            parent  = parent,
            text    = "&New chat",
            tooltip = "Create/join a chat room",
            icon    = "new_chat",
        )
        acts = [a(parent, for_user_id)
                for a in (DirectChat, CreateRoom, JoinRoom)]
        self.setMenu(menu.Menu(parent, acts))


class DirectChat(Action):
    def __init__(self, parent: QWidget, for_user_id: str = "") -> None:
        super().__init__(
            parent   = parent,
            text     = "&Direct chat",
            tooltip  = "Start a direct chat with another user",
            icon     = "direct_chat",
            shortcut = "Ctrl+Shift+D" if not for_user_id else "",
        )
        self.for_user_id = for_user_id
        self.setDisabled(True)

class CreateRoom(Action):
    def __init__(self, parent: QWidget, for_user_id: str = "") -> None:
        super().__init__(
            parent   = parent,
            text     = "&Create room",
            tooltip  = "Create a new chat room",
            icon     = "create_room",
            shortcut = "Ctrl+Shift+C" if not for_user_id else "",
        )
        self.for_user_id = for_user_id

    def on_trigger(self, checked: bool) -> None:
        dialogs.CreateRoom(self.for_user_id).open_modeless()

class JoinRoom(Action):
    def __init__(self, parent: QWidget, for_user_id: str = "") -> None:
        super().__init__(
            parent   = parent,
            text     = "&Join room",
            tooltip  = "Join an existing chat room",
            icon     = "join_room",
            shortcut = "Ctrl+Shift+J" if not for_user_id else "",
        )
        self.for_user_id = for_user_id
        self.setDisabled(True)

class InviteToRoom(Action):
    def __init__(self, parent: QWidget, room: Room, as_user: str = "") -> None:
        tooltip = "Invite a user to join selected room"
        super().__init__(
            parent              = parent,
            text                = "&Invite to room",
            icon                = "invite",
            tooltip             = tooltip,
            multiselect_text    = "&Invite to selected rooms",
            multiselect_tooltip = tooltip.replace("room", "rooms")
        )
        self.room    = room
        self.as_user = as_user

    def on_trigger(self, checked: bool) -> None:
        dialogs.InviteToRoom(self.room, self.as_user).open_modeless()

class LeaveRoom(Action):
    def __init__(self, parent: QWidget, room: Room,
                 immediate_func: ImmediateFunc = None) -> None:
        tooltip = "Leave and remove selected room from the list"
        super().__init__(
            parent              = parent,
            text                = "&Leave room",
            icon                = "leave",
            tooltip             = tooltip,
            multiselect_text    = "&Leave selected rooms",
            multiselect_tooltip = tooltip.replace("room", "rooms"),
            immediate_func      = immediate_func,
        )
        self.room = room

    def on_trigger(self, checked: bool) -> None:
        try:
            self.room.leave()
        except KeyError:  # matrix_client bug
            pass

class AcceptInvite(Action):
    def __init__(self, parent: QWidget, room: Room,
                 immediate_func: ImmediateFunc = None) -> None:
        tooltip = "Accept invitation and join selected room"
        super().__init__(
            parent              = parent,
            text                = "&Accept invitation",
            icon                = "accept_small",
            tooltip             = tooltip,
            multiselect_text    = "&Accept invitations for selected rooms",
            multiselect_tooltip = tooltip.replace("room", "rooms"),
            immediate_func      = immediate_func,
        )
        self.room = room

    def on_trigger(self, checked: bool) -> None:
        try:
            self.room.client.join_room(self.room.room_id)
        except MatrixRequestError as err:
            data = json.loads(err.content)
            if data["errcode"] == "M_UNKNOWN":
                print("Room gone, error box not implemented")

class DeclineInvite(Action):
    def __init__(self, parent: QWidget, room: Room,
                 immediate_func: ImmediateFunc = None) -> None:
        tooltip = "Decline invitation for selected room"
        super().__init__(
            parent              = parent,
            text                = "&Decline invitation",
            icon                = "cancel_small",
            tooltip             = tooltip,
            multiselect_text    = "&Decline invitations for selected rooms",
            multiselect_tooltip = tooltip.replace("room", "rooms"),
            immediate_func      = immediate_func,
        )
        self.room           = room
        self._leave         = LeaveRoom(parent, room, immediate_func)

    def on_trigger(self, checked: bool) -> None:
        self._leave.on_trigger(checked)


# Status

class SetStatus(Action):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(
            parent   = parent,
            text     = "&Status",
            tooltip  = "Change status for all accounts",
            icon     = "status_set_tmp_disabled",
        )
        acts = [a(parent) for a in (Online, Away, Invisible, Offline)]
        self.setMenu(menu.Menu(parent, acts))
        self.setDisabled(True)

class StatusAction(Action):
    def __init__(self, parent, text = None, shortcut = None) -> None:
        name = type(self).__name__
        super().__init__(
            parent   = parent,
            text     = text or f"&{name}",
            tooltip  = f"Change status for all accounts to {name.lower()}",
            icon     = f"status_{name.lower()}",
            shortcut = shortcut or f"Ctrl+Alt+{name[0].upper()}",
        )
        self.setDisabled(True)

class Online(StatusAction):
    pass

class Away(StatusAction):
    pass

class Invisible(StatusAction):
    pass

class Offline(StatusAction):
    def __init__(self, parent):
        super().__init__(parent, text="O&ffline", shortcut="Ctrl+Alt+F")


# Accounts

class AddAccount(Action):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(
            parent   = parent,
            text     = "Add &account",
            tooltip  = "Add a new account to Harmony",
            icon     = "account_add",
        )
        acts = [a(parent) for a in (Login, Register)]
        self.setMenu(menu.Menu(parent, acts))

class DelAccount(Action):
    def __init__(self, parent: QWidget, user_id: str) -> None:
        tooltip  = "Remove selected account from Harmony\n" \
                   "The account will still exist on the server and can " \
                   "be added again later"
        super().__init__(
            parent              = parent,
            icon                = "account_del",
            text                = "Remove &account",
            tooltip             = tooltip,
            multiselect_text    = "&Remove selected accounts",
            multiselect_tooltip = tooltip.replace("account", "accounts")
        )
        self.user_id = user_id

    def on_trigger(self, checked: bool) -> None:
        main_window().accounts.remove(self.user_id)

class Login(Action):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(
            parent   = parent,
            text     = "&Login",
            tooltip  = "Login to an existing Matrix account",
            icon     = "login",
            shortcut = "Ctrl+Shift+L",
        )

    def on_trigger(self, checked: bool) -> None:
        dialogs.Login().open_modeless()


class Register(Action):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(
            parent   = parent,
            text     = "&Register",
            tooltip  = "Register a new Matrix account",
            icon     = "register",
            shortcut = "Ctrl+Shift+R",
        )


    def on_trigger(self, checked: bool) -> None:
        webbrowser.open_new_tab("https://riot.im/app")


# Interface

class ToggleTitleBars(Action):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(
            parent   = parent,
            text     = "&Toggle title bars",
            tooltip  = "Toggle showing dock title bars",
            icon     = "ui_view",
            shortcut = "Ctrl+Shift+T",
        )

    def on_trigger(self, checked: bool) -> None:
        main_window().show_dock_title_bars()


# Settings

class Preferences(Action):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(
            parent   = parent,
            text     = "&Preferences",
            tooltip  = "Change preferences",
            icon     = "preferences_tmp_disabled",
            shortcut = "Ctrl+Shift+P",
        )
        self.setDisabled(True)



# file Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import json
from multiprocessing.pool import ThreadPool

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QLabel

from matrix_client.errors import MatrixRequestError

from . import base
from .. import main_window


class CreateRoom(base.GridDialog):
    # User ID, Room ID
    room_created_signal = pyqtSignal(str, str)

    def __init__(self, for_user_id: str = "") -> None:
        super().__init__("Create room")
        self._pool = ThreadPool(8)
        self.room_created_signal.connect(self.on_room_created)

        logged_in = sorted(main_window().accounts.keys())

        self.info_line = base.InfoLine(self)
        self.about     = QLabel(
            "All fields are optional.<br>"
            "Room can be further customized after creation.<br>"
            "<i>Settings with a * cannot be changed later.</i>",
            self
        )
        self.about.setTextFormat(Qt.RichText)
        self.creator   = base.ComboBox(
            self,
            "Create with account:",
            "Select the account that will be the administrator of this room",
            items=logged_in, initial_item=for_user_id
        )
        self.name      = base.Field(self, "Room name:")
        self.invitees  = base.Field(
            self,
            "Users to invite:",
            "@user1:server.tld\n@user2:matrix.org\n…",
            "",
            "Whitespace-delimited list of user IDs to invite in this room",
            lines = 3
        )
        self.public = base.CheckBox(
            self,
            "&Make this room public",
            "Publish this room in the public room list; allow anyone to join."
        )
        self.federate = base.CheckBox(
            self,
            "&Allow users from different servers*",
            ("User accounts from different homeservers than your account's "
             "will be able to join.\n"
             "These servers will get a copy of the room's history.\n"
             "This setting cannot be changed later."),
            check=True
        )
        self.create = base.AcceptButton(self, "&Create", self.validate)
        self.cancel = base.CancelButton(self, "Ca&ncel")

        if for_user_id:
            self.name.text_edit.setFocus()

        blank = lambda: base.BlankLine(self)

        self.add_spacer(0, 0)
        self.add_spacer(0, 3)
        # widget, from row, from col, row span, col span, align = left
        self.grid.addWidget(self.info_line,  1,  1, 1, 2)
        self.grid.addWidget(self.about,      2,  1, 1, 2)
        self.grid.addWidget(blank(),         3,  1, 1, 2)
        self.grid.addWidget(self.creator,    4,  1, 1, 2)
        self.grid.addWidget(self.name,       5,  1, 1, 2)
        self.grid.addWidget(self.invitees,   7,  1, 1, 2)
        self.grid.addWidget(blank(),         8,  1, 1, 2)
        self.grid.addWidget(self.public,     9,  1, 1, 2, Qt.AlignCenter)
        self.grid.addWidget(self.federate,   10, 1, 1, 2, Qt.AlignCenter)
        self.grid.addWidget(blank(),         11, 1, 1, 2)
        self.grid.addWidget(self.create,     12, 1, 1, 1)
        self.grid.addWidget(self.cancel,     12, 2, 1, 1)
        self.add_spacer(13, 1)
        self.add_spacer(13, 3)

        for half_col in (1, 2):
            self.grid.setColumnMinimumWidth(half_col, 160)

        main_window().events.signals.new_account.connect(self.on_new_account)
        main_window().events.signals.account_gone.connect(self.on_account_gone)


    def on_new_account(self, user_id: str) -> None:
        self.creator.combo_box.addItem(user_id)


    def on_account_gone(self, user_id: str) -> None:
        self.creator.del_items_with_text(user_id)


    def validate(self, _) -> None:
        self.info_line.set_info("Creating room...")

        creator  = self.creator.combo_box.currentText()
        name     = self.name.get_text()
        invitees = [t.strip() for t in self.invitees.get_text().split()
                    if t.strip()]
        public   = self.public.isChecked()
        federate = self.federate.isChecked()

        if not creator:
            self.info_line.set_err("No creator account selected")
            return

        try:
            client = main_window().accounts[creator]
        except KeyError:
            self.info_line.set_err("Selected creator not connected")
            return

        def create() -> None:
            # client.create_room doesn't have a federate param
            answer = client.api.create_room(
                name=name, is_public=public, invitees=invitees,
                federate=federate
            )
            # pylint: disable=protected-access
            room = client._mkroom(answer["room_id"])
            client.join_room(room.room_id)
            self.room_created_signal.emit(client.user_id, room.room_id)

        self._pool.apply_async(create, error_callback=self.on_error)


    def on_error(self, err: BaseException) -> None:
        # Without this handler, exceptions will be silently ignored
        if isinstance(err, MatrixRequestError):
            data = json.loads(err.content)
            self.info_line.set_err(data["error"].replace("user_id", "user ID"))
        else:
            raise err


    def on_room_created(self, user_id: str, room_id: str) -> None:
        self.done(0)
        main_window().go_to_chat_dock(user_id, room_id)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import sys
from typing import Callable, List, Optional, Set

from pkg_resources import resource_filename
from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QApplication, QMainWindow

from . import __about__, error_handler
# pylint: disable=redefined-builtin
from .__about__ import __doc__

_APP:               Optional[QApplication]     = None
_MAIN_WINDOW:       Optional[QMainWindow]      = None


def app() -> QApplication:
    if not _APP:
        raise RuntimeError("Application not initialized.")
    return _APP


def main_window() -> QMainWindow:
    if not _MAIN_WINDOW:
        raise RuntimeError("Main window not initialized.")
    return _MAIN_WINDOW


def on_ctrl_c() -> None:
    try:
        main_window().normal_close = True
    except RuntimeError:
        pass


def run(argv: Optional[List[str]] = None) -> None:
    from . import main
    # pylint: disable=global-statement
    global _APP
    _APP = main.App(argv or sys.argv)

    global _MAIN_WINDOW
    _MAIN_WINDOW = main.MainWindow()
    _MAIN_WINDOW.construct()

    # Make CTRL-C work
    timer = QTimer()
    timer.timeout.connect(on_ctrl_c)
    timer.start(100)

    sys.exit(_APP.exec_())



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from PyQt5.QtWidgets import QMessageBox, QWidget

from .. import actions, main_window


class AcceptRoomInvite(QMessageBox):
    def __init__(self, parent: QWidget, room, room_name: str, inviter_id: str
                ) -> None:
        super().__init__(parent)
        self.setStyleSheet(main_window().theme.style("interface"))
        self.setWindowOpacity(0.9)
        self.setWindowTitle("Harmony - Accept room invitation")

        user_id = room.client.user_id
        self.setText(f"<b>{user_id}</b> has been invited to join the "
                     f"room <b>{room_name}</b> by <b>{inviter_id}</b>.")

        self.yes    = self.addButton("Accept",  QMessageBox.YesRole)
        self.no     = self.addButton("Decline", QMessageBox.NoRole)
        self.cancel = self.addButton("Cancel",  QMessageBox.RejectRole)

        self.yes.clicked.connect(actions.AcceptInvite(self, room).on_trigger)
        self.no.clicked.connect(actions.DeclineInvite(self, room).on_trigger)

        self.setDefaultButton(self.yes)
        self.setIcon(QMessageBox.Question)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from typing import Optional

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QMouseEvent
from PyQt5.QtWidgets import QDockWidget, QLabel, QWidget


class Dock(QDockWidget):
    def __init__(self,
                 title:              str,
                 parent:             QWidget,
                 can_hide_title_bar: bool = True,
                 show_title_bar:     bool = True) -> None:
        super().__init__(title, parent)
        self.current_area: Optional[Qt.DockWidgetArea] = None

        self.title_bar:          TitleBar = TitleBar(title, self)
        self.can_hide_title_bar: bool     = can_hide_title_bar
        self.title_bar_shown:    bool     = show_title_bar
        self.show_title_bar(show_title_bar)

        self.title = title

        # When e.g. user select the tab for this dock
        self.visibilityChanged.connect(self.on_visibility_change)
        self.dockLocationChanged.connect(self.on_location_change)


    def show_title_bar(self, show: Optional[bool] = None) -> None:
        if not self.can_hide_title_bar:
            show = True

        if show is None:
            show = not self.title_bar_shown

        self.setTitleBarWidget(self.title_bar if show else QWidget())
        self.title_bar_shown = show


    @property
    def title(self) -> str:
        return self.title_bar.text()

    @title.setter
    def title(self, value: str) -> None:
        self.title_bar.setText(value)
        self.setWindowTitle(value)


    def focus(self) -> None:
        self.show()
        self.raise_()


    def on_visibility_change(self, visible: bool) -> None:
        if visible:
            self.focus()


    def on_location_change(self, new_area: Qt.DockWidgetArea) -> None:
        self.current_area = new_area
        self.focus()


class TitleBar(QLabel):
    def mousePressEvent(self, event: QMouseEvent) -> None:
        super().mousePressEvent(event)

        if event.button() == Qt.MiddleButton:
            self.parent().hide()



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from typing import Callable, Dict, List, Optional, Tuple

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QCloseEvent
from PyQt5.QtWidgets import (
    QApplication, QDesktopWidget, QMainWindow, QTabWidget, QWidget
)

HooksType = Dict[str, Callable[["MainWindow"], None]]

HOOKS_INIT_1_START:        HooksType = {}
HOOKS_INIT_2_BEFORE_LOGIN: HooksType = {}
HOOKS_INIT_3_END:          HooksType = {}

# pylint: disable=wrong-import-position
from . import (
    __about__, app, error_handler, homepage, theming, toolbar, usertree
)
from .accounts import AccountManager
from .chat import ChatDock
from .dock import Dock
from .event_logger import EventLogger
from .events import EventManager
from .shortcuts import Shortcut, ShortcutManager



class App(QApplication):
    def __init__(self, argv: Optional[List[str]] = None) -> None:
        super().__init__(argv or [])
        # if not self.styleSheet:  # user can load one with --stylesheet
            # self.
            # pass

        self.setApplicationName(__about__.__pkg_name__)
        self.setApplicationVersion(__about__.__version__)

        self.last_focused_dock:      Optional[Dock]     = None
        self.last_focused_chat_dock: Optional[ChatDock] = None
        self.focusChanged.connect(self.on_focus_change)


    def on_focus_change(self, _: QWidget, new: QWidget) -> None:
        while not isinstance(new, Dock):
            if new is None:
                return
            new = new.parent()

        self.last_focused_dock = new

        if isinstance(new, ChatDock):
            self.last_focused_chat_dock = new


    def get_focused(self, condition: Optional[Callable[[QWidget], bool]] = None
                   ) -> Optional[Dock]:
        widget = self.focusWidget()

        if not condition:
            return widget

        while widget is not None:
            if condition(widget):
                return widget
            widget = widget.parent()

        return None


class MainWindow(QMainWindow):
    def __init__(self) -> None:
        super().__init__()
        self.title_bars_shown: bool = False
        self.normal_close:     bool = False


    def construct(self) -> None:
        # pylint: disable=attribute-defined-outside-init
        # Can't define all that __init__ instead.
        # The UI elements need _MAIN_WINDOW to be set, see run() in __init__.

        for func in HOOKS_INIT_1_START.values():
            func(self)

        # Setup error console:

        self.error_dock = Dock("Error console", self, can_hide_title_bar=False)
        self.error_dock.hide()
        self.error_dock.setWidget(error_handler.console.Console())


        # Setup appearance:

        self.setWindowTitle("Harmony")
        screen = QDesktopWidget().screenGeometry()
        self.resize(min(screen.width(), 1152), min(screen.height(), 768))

        # self.setAttribute(Qt.WA_TranslucentBackground)
        self.setWindowOpacity(0.9)

        self.theme = theming.Theme("glass")
        self.icons = theming.Icons("flat_white")
        self.setStyleSheet(self.theme.style("interface"))


        # Setup main classes and event listeners:

        self.event_logger: EventLogger     = EventLogger()
        self.accounts:     AccountManager  = AccountManager()
        self.events:       EventManager    = EventManager()
        self.shortcuts:    ShortcutManager = ShortcutManager()

        self.event_logger.start()

        self.events.signals.left_room.connect(self.remove_chat_dock)
        self.events.signals.room_rename.connect(self.update_chat_dock_name)
        # Triggered by room renames that happen when account changes
        # self.events.signal.account_change.connect(self.update_chat_dock_name)


        # Setup main UI parts:

        self.setDockNestingEnabled(True)
        self.setTabPosition(Qt.AllDockWidgetAreas, QTabWidget.North)

        self.tree_dock = Dock("Accounts / Rooms", self)
        self.tree_dock.setWidget(usertree.UserTree())
        self.addDockWidget(Qt.LeftDockWidgetArea, self.tree_dock)

        self.actions_dock = Dock("Actions", self)
        self.actions_dock.setWidget(toolbar.ActionsBar())
        self.splitDockWidget(self.tree_dock, self.actions_dock, Qt.Vertical)

        self.home_dock = Dock("Home", self)
        self.home_dock.setWidget(homepage.HomePage())
        self.addDockWidget(Qt.RightDockWidgetArea, self.home_dock)

        # {(user_id, room_id): dock}
        self.visible_chat_docks: Dict[Tuple[str, str], ChatDock] = {}

        def on_activation() -> None:
            try:
                app().last_focused_chat_dock.focus()
            except AttributeError:
                pass

        self.shortcuts.add(Shortcut(
            name          = "Focus last chat",
            on_activation = on_activation,
            bindings      = ["A-c"],
            autorepeat    = False
        ))

        self.show_dock_title_bars(False)

        # Run:

        self.show()

        for func in HOOKS_INIT_2_BEFORE_LOGIN.values():
            func(self)

        try:
            self.accounts.login_using_config()
        except FileNotFoundError:
            pass

        for func in HOOKS_INIT_3_END.values():
            func(self)


    def show_dock_title_bars(self, show: Optional[bool] = None) -> None:
        if show is None:
            show = not self.title_bars_shown

        docks = (self.tree_dock, self.actions_dock, self.home_dock,
                 self.error_dock, *self.visible_chat_docks.values())

        for dock in docks:
            dock.show_title_bar(show)

        self.title_bars_shown = show


    def show_error_dock(self) -> None:
        if self.error_dock.isVisible():
            return

        self.error_dock.widget().display.apply_style()

        try:
            tab_with = app().last_focused_chat_dock
            if not tab_with or not tab_with.isVisible():
                tab_with = self.home_dock
        except AttributeError:
            self.addDockWidget(Qt.RightDockWidgetArea, self.error_dock)
        else:
            self.tabifyDockWidget(tab_with, self.error_dock)


    def new_chat_dock(self,
                      user_id:            str,
                      room_id:            str,
                      previously_focused: ChatDock,
                      in_new:             str            = "",
                      split_orientation:  Qt.Orientation = Qt.Horizontal,
                     ) -> ChatDock:
        assert in_new in (None, "", "tab", "split")

        prev_is_in_tab = bool(self.tabifiedDockWidgets(previously_focused))
        dock = ChatDock(user_id, room_id, self)

        if not previously_focused.isVisible():
            self.addDockWidget(
                previously_focused.current_area or Qt.RightDockWidgetArea,
                dock,
                split_orientation
            )

        if in_new == "split" and prev_is_in_tab:
            area = None if previously_focused.isFloating() else \
                   previously_focused.current_area
            area = area or Qt.RightDockWidgetArea

            self.addDockWidget(area, dock, split_orientation)

        elif in_new == "split":
            self.splitDockWidget(previously_focused, dock, split_orientation)

        else:
            self.tabifyDockWidget(previously_focused, dock)

        return dock


    def go_to_chat_dock(self, user_id: str, room_id: str, in_new: str = "",
                        split_orientation: Qt.Orientation = Qt.Horizontal
                       ) -> None:

        dock = self.visible_chat_docks.get((user_id, room_id))
        if dock:
            dock.focus()
            return

        dock = app().last_focused_chat_dock
        if not dock or not dock.isVisible():
            dock = self.home_dock

        if in_new:
            dock = self.new_chat_dock(user_id, room_id, dock, in_new,
                                      split_orientation)
        elif isinstance(dock, ChatDock):
            dock.change_room(user_id, room_id)
        else:
            self.home_dock.hide()
            dock = self.new_chat_dock(user_id, room_id, dock, in_new,
                                      split_orientation)

        dock.focus()


    def remove_chat_dock(self, user_id: str, room_id: str) -> None:
        dock = self.visible_chat_docks.get((user_id, room_id))
        if dock:
            dock.hide()


    def update_chat_dock_name(self, user_id: str, room_id: str) -> None:
        dock = self.visible_chat_docks.get((user_id, room_id))
        if dock:
            dock.autoset_title()


    def closeEvent(self, event: QCloseEvent) -> None:
        self.normal_close = True
        super().closeEvent(event)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from typing import List

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import (
    QSizePolicy, QToolBar, QToolButton, QWidget, QWidgetAction
)

from . import __about__, actions, main_window, menu


class ActionsBar(QToolBar):
    button_visibility_changed_signal = pyqtSignal()

    def __init__(self) -> None:
        super().__init__(main_window())
        self.buttons: List[ActionButton] = []

        self.expand_button: QToolButton = self.children()[1]
        self.expand_button.setIcon(main_window().icons.icon("toolbar_expand"))
        self.expand_button.setProperty("is-expand-button", True)
        self.update_expand_menu()
        self.button_visibility_changed_signal.connect(self.update_expand_menu)

        acts = [actions.SetStatus, actions.AddAccount, actions.NewChat,
                actions.ToggleTitleBars, actions.Preferences]

        for action in acts:
            actwa = ActionWidget(self, action(main_window()))
            self.addAction(actwa)
            self.buttons.append(actwa.button)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)


    def update_expand_menu(self) -> None:
        # We need to handle this menu ourselves, else it will be made of
        # ActionButtons and not be a proper menu with shortcuts, submenus, etc.
        acts = [b.action for b in self.buttons if not b.isVisible()]
        self.expand_button.setMenu(menu.Menu(self, acts))


class ActionWidget(QWidgetAction):
    def __init__(self, bar: ActionsBar, action: actions.Action) -> None:
        super().__init__(bar)
        self.bar    = bar
        self.action = action
        self.button = ActionButton(self.bar, self.action)


    def createWidget(self, _: QWidget) -> int:
        return self.button

    @staticmethod
    def deleteWidget(widget: QWidget) -> None:
        # Original method deletes it from memory, causing expand menu to crash
        widget.hide()


class ActionButton(QToolButton):
    def __init__(self, bar: ActionsBar, action: actions.Action) -> None:
        super().__init__(bar)
        self.bar    = bar
        self.action = action

        self.setDefaultAction(self.action)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setPopupMode(QToolButton.InstantPopup)
        self.setToolButtonStyle(Qt.ToolButtonIconOnly)


    def showEvent(self, event) -> None:
        super().showEvent(event)
        self.bar.button_visibility_changed_signal.emit()

    def hideEvent(self, event) -> None:
        super().hideEvent(event)
        self.bar.button_visibility_changed_signal.emit()



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import re
import subprocess as sp
from typing import Union

from . import register, say, utils
from ..chat import Chat


@register
def shell(chat: Chat, args: dict) -> None:
    """
    Usage:
      /shell COMMAND [-e|--echo -c|--hide-cmd -b|--no-block]

    Run a `COMMAND` with the OS shell, send standard output.

    When the command finishes,
    - Its return code will be shown if it isn't `0`
    - stderr will be shown if it contains anything
    - stdout will be shown/sent if the command outputted anything.

    ANSI escape codes are stripped from outputs.

    Options:
      -e, --echo
        Print standard output locally without sending it.

      -c, --hide-cmd
        Do not print the command used before output.

      -b, --no-block
        Do not format the output as a monospace font code block.

    Examples (POSIX):
    ```
        /shell uptime
        /shell date --hide-cmd --no-block
        /shell 'echo "It is $(date '+%T') here"' -cb
        /shell "ls -l ~"
        /shell "echo $SHELL"
        /shell "neofetch || screenfetch && echo 'It works!'"
    ```"""

    shell_f(
        chat          = chat,
        command       = args["COMMAND"],
        echo          = args["--echo"],
        no_command    = args["--hide-cmd"],
        no_code_block = args["--no-block"],
    )


def shell_f(chat:          Chat,
            command:       str,
            echo:          bool          = False,
            no_command:    bool          = False,
            no_code_block: bool          = False) -> None:

    process = sp.Popen(
        command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE,
    )
    stdout, stderr = process.communicate()
    retcode        = process.returncode

    command = _treat_output(command)
    stdout  = _treat_output(stdout)
    stderr  = _treat_output(stderr)

    if retcode and retcode != 0:
        utils.print_warn(chat, f"`{command}` returned `{retcode}`.")

    if stderr:
        utils.print_warn(chat, f"```\n{stderr}\n```")

    if stdout:
        bticks = ("", "") if no_code_block else ("```\n", "\n```")
        text   = "\n".join((
            bticks[0] if no_command else f"{bticks[0]}$ {command}",
            f"{stdout}{bticks[1]}"
        ))

        if echo:
            utils.print_info(chat, text)
        else:
            say.say_f(chat, text)


def _treat_output(text: Union[bytes, str]) -> str:
    if isinstance(text, bytes):
        text = text.decode("utf-8")

    # Strip ANSI escape sequences:
    text = re.sub(r"\x1B\[[0-?]*[ -/]*[@-~]", "", text)

    # zero-width space to "escape" triple-backticks and prevent glitching
    # the code block/command without altering what the user sees or can select:
    text = text.replace("```", "\u200b```")

    # Strip whitespace at the end
    return text.rstrip()



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import base64
import hashlib
import json
import os
import platform
from collections import UserDict
from multiprocessing.pool import ThreadPool
import threading
from typing import Dict, List, Optional
from urllib.parse import urlparse

from atomicfile import AtomicFile
from PyQt5.QtCore import QObject, QStandardPaths, pyqtSignal

from matrix_client.client import MatrixClient

LOAD_NUM_EVENTS_ON_START = 10


class _SignalObject(QObject):
    # Signals can only be emited from QObjects, but AccountManager
    # can't inherit from it because of metaclass conflict.
    login  = pyqtSignal(MatrixClient)
    logout = pyqtSignal(str)


class AccountManager(UserDict):
    def __init__(self) -> None:
        super().__init__()
        self.signals = _SignalObject()
        self._pool   = ThreadPool(8)
        self._lock   = threading.Lock()

    # Login/logout

    def login_using_config(self, path: str = "") -> None:
        def log(acc: Dict[str, str]) -> None:
            self.login(acc["server_url"], acc["user_id"], acc["password"],
                       acc["device_name"])

        def err_c(err: BaseException) -> None:
            raise err

        self._pool.map_async(log, self.config_read(path), error_callback=err_c)


    def login(self,
              server_url:    str,
              user_id:       str,
              password:      str  = "",
              device_name:   str  = "",
              add_to_config: bool = False) -> None:

        user_id = user_id.strip()

        if user_id.startswith("@") and ":" not in user_id:
            user_id = user_id[1:]

        if not user_id.startswith("@"):
            user_id = f"@{user_id}:{urlparse(server_url).netloc}"

        if user_id in self.data:
            return

        db_path, db_filename = os.path.split(self.get_crypt_db_path(user_id))

        client = MatrixClient(
            server_url,
            sync_filter_limit = LOAD_NUM_EVENTS_ON_START,
            encryption        = True,
            restore_device_id = True,
            encryption_conf   = {
                "store_conf": {"db_path": db_path, "db_name": db_filename}
            }
        )

        client.login(user_id, password, sync=False)

        self.data[user_id] = client
        self.signals.login.emit(client)

        system = f" on {platform.system()}".rstrip()
        system = f"{system} {platform.release()}".rstrip() \
                 if system != " on" else ""

        client.api.update_device_info(client.device_id,
                                      device_name or f"Harmony{system}")

        if add_to_config:
            self.config_add(server_url, user_id, password, device_name)


    def remove(self, user_id: str) -> None:
        if user_id not in self.data:
            return

        self.signals.logout.emit(user_id)
        self._pool.apply_async(self.data[user_id].logout)
        del self.data[user_id]
        self.config_del(user_id)


    # Standard file paths

    @staticmethod
    def _get_standard_path(kind:            QStandardPaths.StandardLocation,
                           file:            str,
                           initial_content: Optional[str] = None) -> str:
        relative_path = file.replace("/", os.sep)

        path = QStandardPaths.locate(kind, relative_path)
        if path:
            return path

        base_dir = QStandardPaths.writableLocation(kind)
        path     = f"{base_dir}{os.sep}{relative_path}"
        os.makedirs(os.path.split(path)[0], exist_ok=True)

        if initial_content is not None:
            with AtomicFile(path, "w") as new:
                new.write(initial_content)

        return path


    @property
    def standard_accounts_config_path(self) -> str:
        return self._get_standard_path(
            QStandardPaths.AppConfigLocation, "accounts.json", "[]"
        )


    def get_crypt_db_path(self, user_id: str) -> str:
        safe_filename = hashlib.md5(user_id.encode("utf-8")).hexdigest()
        return self._get_standard_path(
            QStandardPaths.AppDataLocation, f"encryption/{safe_filename}.db"
        )


    # Config file operations

    def config_read(self, path: str = "") -> List[Dict[str, str]]:
        with open(path or self.standard_accounts_config_path, "r") as file:
            accs = json.loads(file.read().strip()) or []

        for acc in accs:
            pw = acc["password"]
            acc["password"] = \
                str(base64.b85decode(base64.b64decode(pw))[::-1], "utf-8")

        return accs


    def config_add(self, server_url: str, user_id: str, password: str,
                   device_name: str = "", path: str = "") -> None:
        with self._lock:
            path     = path or self.standard_accounts_config_path
            accounts = self.config_read(path)
            params   = {"server_url": server_url, "user_id":     user_id,
                        "password":   password,   "device_name": device_name,}

            for i, acc in enumerate(accounts):
                if acc["user_id"] == user_id:
                    if acc == params:
                        return

                    accounts[i] = params
                    break
            else:
                accounts.append(params)

            with AtomicFile(path, "w") as new:
                new.write(self._serialize_config(accounts))


    def config_del(self, user_id: str, path: str = "") -> None:
        with self._lock:
            path     = path or self.standard_accounts_config_path
            accounts = self.config_read(path)
            accounts = [a for a in accounts if a["user_id"] != user_id]

            with AtomicFile(path, "w") as new:
                new.write(self._serialize_config(accounts))


    @staticmethod
    def _serialize_config(accounts: List[Dict[str, str]]) -> str:
        for acc in accounts:
            pw  = acc["password"]
            byt = base64.b64encode(base64.b85encode(bytes(pw[::-1], "utf-8")))

            acc["password"] = str(byt, "utf-8")

        return json.dumps(accounts, indent=4, ensure_ascii=False)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from multiprocessing.pool import ThreadPool

from PyQt5.QtCore import Qt, pyqtSignal

from matrix_client.errors import (
    MatrixError, MatrixHttpLibError, MatrixRequestError
)

from . import base
from .. import main_window


class Login(base.GridDialog):
    login_done_signal = pyqtSignal()

    def __init__(self) -> None:
        super().__init__("Login")
        self._pool = ThreadPool(1)

        self.info_line = base.InfoLine(self)
        self.server    = base.Field(
            self, "Server:", "scheme://domain.tld", "https://matrix.org"
        )
        self.username = base.Field(self, "Username:")
        self.password = base.Field(self, "Password:", is_password=True)
        self.remember = base.CheckBox(
            self,
            "&Remember this account",
            "Automatically login this account on startup",
            check = True,
        )
        self.login = base.AcceptButton(
            self, "&Login", self.validate,
            (self.server, self.username, self.password),
        )
        self.cancel = base.CancelButton(self)

        self.username.text_edit.setFocus()

        blank = lambda: base.BlankLine(self)

        self.add_spacer(0, 0)
        self.add_spacer(0, 3)
        # widget, from row, from col, row span, col span, align = left
        self.grid.addWidget(self.info_line,  1, 1, 1, 2)
        self.grid.addWidget(self.server,     2, 1, 1, 2)
        self.grid.addWidget(self.username,   3, 1, 1, 2)
        self.grid.addWidget(self.password,   4, 1, 1, 2)
        self.grid.addWidget(blank(),         5, 1, 1, 2)
        self.grid.addWidget(self.remember,   6, 1, 1, 2, Qt.AlignCenter)
        self.grid.addWidget(blank(),         7, 1, 1, 2)
        self.grid.addWidget(self.login ,     8, 1, 1, 1)
        self.grid.addWidget(self.cancel,     8, 2, 1, 1)
        self.add_spacer(9, 1)
        self.add_spacer(9, 3)

        for half_col in (1, 2):
            self.grid.setColumnMinimumWidth(half_col, 144)

        self.login_done_signal.connect(self.on_login)


    def validate(self, _) -> None:
        self.info_line.set_info("Logging in...")

        self._pool.apply_async(
            func = main_window().accounts.login,
            kwds = {
                "server_url":    self.server.get_text(),
                "user_id":       self.username.get_text(),
                "password":      self.password.get_text(),
                "add_to_config": self.remember.isChecked(),
            },
            callback       = lambda *_: self.login_done_signal.emit(),
            error_callback = self.on_error,
        )


    def on_login(self) -> None:
        self.done(0)


    def on_error(self, err: BaseException) -> None:
        if isinstance(err, MatrixRequestError):
            self.info_line.set_err("Invalid username or password")

        elif isinstance(err, MatrixHttpLibError):
            self.info_line.set_err("Unreachable server")

        elif isinstance(err, MatrixError):
            if "No scheme" in str(err):
                self.info_line.set_err("Server is missing a scheme://")

            elif "Invalid home server" in str(err):
                self.info_line.set_err("Invalid server URL")

            else:
                raise err

        else:
            raise err



# file Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import json
from multiprocessing.pool import ThreadPool
from typing import Set

from PyQt5.QtCore import pyqtSignal

from matrix_client.errors import MatrixRequestError
from matrix_client.room import Room

from . import base
from .. import main_window


class InviteToRoom(base.GridDialog):
    invites_sent_signal = pyqtSignal()


    def __init__(self, room: Room, as_user: str = "") -> None:
        super().__init__()
        self.room  = room
        self._pool = ThreadPool(8)
        self.update_wintitle()

        us_in_room = {i for i in main_window().accounts if i in room.members}

        self.info_line = base.InfoLine(self)
        self.sender    = base.ComboBox(
            self,
            "Send invite as:",
            "Select the account that will be used to send this invite",
            items=sorted(us_in_room), initial_item=as_user
        )
        self.invitees  = base.Field(
            self,
            "Users to invite:",
            "@user1:server.tld\n@user2:matrix.org\n…",
            "",
            "Whitespace-delimited list of user IDs to invite in this room",
            lines = 9
        )
        self.send   = base.AcceptButton(self, "&Send", self.validate)
        self.cancel = base.CancelButton(self, "&Cancel")

        if as_user:
            self.invitees.text_edit.setFocus()

        blank = lambda: base.BlankLine(self)

        self.add_spacer(0, 0)
        self.add_spacer(0, 3)
        # widget, from row, from col, row span, col span, align = left
        self.grid.addWidget(self.info_line,  1, 1, 1, 2)
        self.grid.addWidget(self.sender,     2, 1, 1, 2)
        self.grid.addWidget(self.invitees,   3, 1, 1, 2)
        self.grid.addWidget(blank(),         4, 1, 1, 2)
        self.grid.addWidget(self.send,       5, 1, 1, 1)
        self.grid.addWidget(self.cancel,     5, 2, 1, 1)
        self.add_spacer(6, 1)
        self.add_spacer(6, 3)

        for half_col in (1, 2):
            self.grid.setColumnMinimumWidth(half_col, 160)

        self.invites_sent_signal.connect(self.on_invites_sent)
        main_window().events.signals.new_account.connect(self.on_new_login)
        main_window().events.signals.account_gone.connect(self.on_account_gone)
        main_window().events.signals.room_rename.connect(self.update_wintitle)


    def update_wintitle(self) -> None:
        self.setWindowTitle(
            f"Harmony - Invite to room: {self.room.display_name}"
        )


    def on_new_login(self, user_id: str) -> None:
        for member in self.room.get_joined_members():
            if member.user_id == user_id:
                self.creator.combo_box.addItem(user_id)


    def on_account_gone(self, user_id: str) -> None:
        self.sender.del_items_with_text(user_id)


    def validate(self, _) -> None:
        self.info_line.set_info("Sending invites...")

        sender   = self.sender.combo_box.currentText()
        invitees = self.invitees.get_text().split()

        if not sender:
            self.info_line.set_err("No sender account selected")
            return

        try:
            client = main_window().accounts[sender]
        except KeyError:
            self.info_line.set_err("Selected sender not connected")
            return

        sent: Set[str] = set()

        def send(to_user_id: str) -> None:
            # If user gets an error, edit the invitees and click "Send" again:
            if to_user_id in sent:
                return
            # client.invite_user won't let us handle errors
            client.api.invite_user(self.room.room_id, to_user_id)
            sent.add(to_user_id)

        def done(_) -> None:
            self.invites_sent_signal.emit()

        self._pool.map_async(send, invitees,
                             callback=done, error_callback=self.on_error)


    def on_error(self, err: BaseException) -> None:
        # Without this handler, exceptions will be silently ignored
        if isinstance(err, MatrixRequestError):
            data = json.loads(err.content)
            self.info_line.set_err(data["error"])
        else:
            raise err


    def on_invites_sent(self) -> None:
        self.done(0)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import threading

from . import register, utils
from ..chat import Chat
from ..message import Message

SEND_LOCK = threading.Lock()


@register
def say(chat: Chat, args: dict) -> None:
    """Usage: /say MESSAGES... [-h|--html -e|--echo]

    Print/send markdown or HTML messages.

    Options:
      -h, --html
        Parse `MESSAGES` as HTML instead of markdown.

      -e, --echo
        Print `MESSAGES` locally without sending them.

    Examples:
    ```
      /say "Same as sending a message without using a command"
      /say "**Bold message 1**" "*Italic message 2*"
      /say "<a href='https://example.com'>Example link</a>" --html
      /say "Talking to myself" --echo
    ```"""

    for msg in args["MESSAGES"]:
        say_f(chat=chat, text=msg, is_html=args["--html"], echo=args["--echo"])


def say_f(chat: Chat, text: str, is_html: bool = False, echo: bool = False
         ) -> None:
    if not text.strip():
        return

    if echo:
        utils.print_info(chat, text, is_html=is_html)
        return


    msg = Message(
        sender_id = chat.client.user_id,
        room_id   = chat.room.room_id,
        markdown  = text if not is_html else "",
        html      = text if is_html     else "",
    )

    # Prevent mixing up if user sends msg B before msg A is received
    # by the server:
    with SEND_LOCK:
        msg.send()



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import re

import markdown2
from markdownify import markdownify

CONVERT_TO_MD_EXTRAS = [
    # Insert line break on \n instead of required two spaces at line end:
    "break-on-newline",
    # Disable _ and __ for italic/bold, only leave * and **:
    "code-friendly",
    # Allow lists to be after a paragraph without an empty line between them:
    "cuddled-lists",
    # Support GitHub ``` code blocks (needs pygments for syntax highlighting):
    "fenced-code-blocks",
    # Parse lines starting with `>>>` as Python interpreter code:
    "pyshell",
    # Parse ~~text~~ as strikethrough
    "strike",
    # Allow GitHub tables syntax, e.g. `a | b | c\n--- | --- | ---\nx | y | z`:
    "tables",
    # Require `# Titles` to have a space between the # and text:
    "tag-friendly",
    # Allow GitHub-style task lists, e.g. `- [x] thing`:
    "task_list",
    # Allow Google Code Wiki tables syntax, e.g. `a || b || c\nx || y || z`:
    "wiki-tables",
]

CONVERT_TO_MD_DISABLED_EXTRAS = [
    # Allow usinbg a `markdown="1"` attribute in block HTML tags to process the
    # text inside as markdown.
    # The Markdown cannot be on the same line as the block HTML element.
    "markdown-in-html",
    # Parse lines starting with `>!` as spoilers, similar to Stack Overflow:
    # Impossible to make it cross-client since Riot doesn't parse span style
    "spoiler",
    # Support foot notes, e.g. `Some text, see [^other-thing]`;
    # (bottom) `[^other-thing]: Description`:
    "footnotes",
    # Generate id attributes for titles,
    # e.g. `# My Title` → `<h1 id="my-title">My Title</h1>`:
    "header-ids",
    # Enable metadata parsing,
    # see https://github.com/trentm/python-markdown2/wiki/metadata
    "metadata",
    # See https://github.com/trentm/python-markdown2/wiki/numbering
    "numbering",
    # Generate a table of content for titles, experimental:
    "toc",
]

_TO_MARKDOWN = markdown2.Markdown(extras    = CONVERT_TO_MD_EXTRAS,
                                  safe_mode = "escape")


def from_html(html: str) -> str:
    return markdownify(html)


def to_html(markdown: str) -> str:
    html = _TO_MARKDOWN.convert(markdown)

    # Apply a class to \> escaped quotes
    html = re.sub(r"(<p>|<br */?>\s*)(>.*?)(?=\s*</p>|<br */?>)",
                  r"\1<span class=escaped-quote>\2</span>",
                  html,
                  flags=re.DOTALL)

    # Qt only knows <s> for striketrough, replace <del> and <strike>
    html = re.sub(r"(</?)\s*(del|strike)>", r"\1s>", html)
    return html



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import abc
from pathlib import Path
from typing import Any, Dict, List, Sequence, Tuple

from pkg_resources import resource_filename
from PyQt5.QtGui import QIcon

from . import __about__

_PKG = __about__.__name__


class ResourceDir(abc.ABC):
    def __init__(self, base_dir: str, res_name: str) -> None:
        self.name:      str  = res_name
        self.dir:       str  = f"{base_dir}/{res_name}"
        self.base_path: Path = Path(resource_filename(_PKG, self.dir))

        # {relative path: {ext: (path, file content as bytes)}}
        self._cache: Dict[str, Dict[str, Tuple[Path, Any]]] = {}
        self.reload()


    def reload(self) -> None:
        self._cache_dir(self.base_path)


    def _cache_dir(self, path: Path) -> None:
        for item in path.iterdir():
            if item.is_dir():
                self._cache_dir(item)
            if item.is_file():
                self._cache_file(item)


    def _cache_file(self, path: Path) -> None:
        key, ext = "/".join(path.relative_to(self.base_path).parts)\
                   .split(".", maxsplit=1)

        file_dict      = self._cache.setdefault(key, {})
        file_dict[ext] = (path, self.cache_file_content(path))


    @abc.abstractmethod
    def cache_file_content(self, path: Path) -> Any:
        pass


    def get_resource(self, relative_path: str, exts: Sequence[str]
                    ) -> Tuple[Path, Any]:
        not_found: List[Path] = []

        for ext in exts:
            try:
                return self._cache[relative_path][ext]
            except KeyError:
                not_found.append(self.base_path / f"{relative_path}.{ext}")

        obj   = type(self).__name__.lower()
        files = "\n".join((f"  {p!s}" for p in not_found))
        raise FileNotFoundError(
            f"None of these files found for {obj} {self.name!r}:\n{files}"
        )


class Theme(ResourceDir):
    def __init__(self, name: str) -> None:
        super().__init__("themes", name)


    def cache_file_content(self, path: Path) -> str:
        return path.read_text()


    def style(self, relative_path: str) -> str:
        return self.get_resource(relative_path, ("qss", "css"))[1]


class Icons(ResourceDir):
    def __init__(self, name: str) -> None:
        super().__init__("icons", name)


    def cache_file_content(self, path: Path) -> QIcon:
        return QIcon(str(path))


    def path(self, relative_path: str) -> str:
        return str(self.get_resource(relative_path, ("png",))[0])


    def icon(self, relative_path: str) -> str:
        return self.get_resource(relative_path, ("png",))[1]



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import json
import os
from typing import List

from atomicfile import AtomicFile

from . import register, utils
from ..chat import CHAT_INIT_HOOKS, Chat
from ..utils import get_config_path
from .eval import eval_f


DEFAULT_CONFIG = """[
    "/alias /h /help",
    "/alias /echo '/say --echo'"
]"""


@register
def autorun(chat: Chat, args: dict) -> None:
    r"""Usage: /autorun [COMMANDS]... [-r|--reload | --reset]

    Add commands to automatically run when Harmony starts.

    If no `COMMANDS` is passed, the `autorun.json` file path will be printed.
    Commands can be manually added, edited or removed in this file.

    Options:
      -r, --reload
        Reload `autorun.json` and execute its commands.
        Note that if you remove a command that had a durable effect like
        `/alias`, the effect will persist until Harmony is restarted.

      --reset
        Reset `autorun.json` back to default and reload. Cannot be undone!

    Examples:
    ```
      /autorun -r
      /autorun "/alias /h /help"
    ```"""

    if args["--reset"]:
        reset_autorun_json(chat)
    elif args["--reload"]:
        load_autorun_json(chat)

    if not args["COMMANDS"]:
        utils.print_info(chat, get_autorun_json_path())
        return

    for cmd in args["COMMANDS"]:
        add_command(chat, cmd)


def get_autorun_json_path() -> str:
    return get_config_path("autorun.json", DEFAULT_CONFIG)


def load_autorun_json(chat: Chat) -> None:
    with open(get_autorun_json_path(), "r") as in_file:
        config: List[str] = json.loads(in_file.read())

    assert isinstance(config, list), \
           "Content of `autorun.json` must be a JSON list of strings."

    for cmd in config:
        eval_f(chat, cmd)


def add_command(chat, command: str) -> None:
    path = get_autorun_json_path()

    with open(path, "r") as in_file, AtomicFile(path, "w") as out_file:
        config: List[str] = json.loads(in_file.read())
        config.append(command)
        out_file.write(json.dumps(config, indent=4, ensure_ascii=False))

    load_autorun_json(chat)


def reset_autorun_json(chat) -> None:
    path = get_autorun_json_path()
    if os.path.exists(path):
        os.remove(path)

    load_autorun_json(chat)


CHAT_INIT_HOOKS["load_autorun_json"] = load_autorun_json



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import base64
import json
import os
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence

from atomicfile import AtomicFile
from dataclasses import dataclass, field
from PyQt5.QtCore import QDateTime, QStandardPaths

from . import main_window


@dataclass
class EventLogger:
    base_dir:                Optional[os.PathLike] = None
    filename_format:         str                   = "{room_id}/{date}.json"
    date_format:             str                   = "yyyy-MM-dd"
    base32_room_id:          bool                  = True
    allow_event_overwriting: bool                  = True

    json_dumps_kwargs: Dict[str, Any] = field(default_factory=dict)

    _pool: ThreadPool = field(init=False, repr=False, default=ThreadPool(1))


    def __post_init__(self) -> None:
        qsp = QStandardPaths
        self.base_dir = (
            Path(self.base_dir) if self.base_dir else
            Path(qsp.writableLocation(qsp.AppDataLocation)) / "logs"
        )
        self.json_dumps_kwargs = self.json_dumps_kwargs or {
            "indent":       4,
            "ensure_ascii": False,
            "sort_keys":    True,
        }
        self._pool = ThreadPool(1)


    def start(self, autolog_funcs: Sequence[Callable[[dict], None]] = (),
             ) -> None:
        for func in autolog_funcs or (self.log_to_file,):
            main_window().events.signals.new_unique_event.connect(
                lambda _, ev, f=func: self._pool.apply_async(
                    f, (ev,), error_callback=self.on_log_error,
                )
            )


    @staticmethod
    def on_log_error(err: BaseException) -> None:
        raise err


    def log_to_file(self, event: dict) -> None:
        path = self.base_dir / self.filename_format.format(  # type: ignore
            room_id = self.encode_room_id(event["room_id"]),
            date    = QDateTime.fromMSecsSinceEpoch(event["origin_server_ts"])\
                      .toString(self.date_format)
        )

        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text("[]")

        with open(path, "r") as in_file, AtomicFile(path, "w") as out_file:
            logged: List[dict] = json.loads(in_file.read())

            if self.allow_event_overwriting:
                logged = [e for e in logged
                          if e["event_id"] != event["event_id"]]
            else:
                for logged_ev in logged:
                    if logged_ev["event_id"] == event["event_id"]:
                        return

            logged.append(event)
            logged.sort(key=lambda ev: ev["origin_server_ts"])

            out_file.write(json.dumps(logged, **self.json_dumps_kwargs))


    def encode_room_id(self, room_id: str) -> str:
        return (
            base64.b32encode(bytes(room_id, "utf-8")).decode("utf-8").lower()
            if self.base32_room_id else room_id
        )


    def decode_room_id(self, encoded: str) -> str:
        return (
            base64.b32decode(bytes(encoded, "utf-8"), casefold=True)
            .decode("utf-8")

            if self.base32_room_id else encoded
        )



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from typing import Optional

from . import register
from ..chat import Chat


@register
def nick(chat: Chat, args: dict) -> None:
    """Usage: /nick [NAME] [-r|--room]

    Set or clear display name, globally or only for this room.

    Options:
      -r, --room
        Set name only for the current room.

    Examples:
    ```
      /nick
      /nick Alice
      /nick "Not Alice" --room
    ```"""

    nick_f(chat=chat, name=args["NAME"], for_room=args["--room"])


def nick_f(chat: Chat, name: Optional[str] = None, for_room: bool = False
          ) -> None:
    if for_room:
        chat.room.set_user_profile(displayname=name or chat.client.user_id)
    else:
        chat.client.user.set_display_name(name or chat.client.user_id)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import time
from threading import Thread
from typing import Deque, Tuple

from PyQt5.QtCore import QDateTime, Qt, pyqtSignal
from PyQt5.QtGui import QKeyEvent, QTextCursor, QTextLength, QTextTableFormat

from . import Chat
from .. import message_display
from ..message import Message


class ChatMessageDisplay(message_display.MessageDisplay):
    _add_message_request = pyqtSignal(Message)


    def __init__(self, chat: Chat) -> None:
        super().__init__()
        self.apply_style()
        self.chat: Chat = chat

        constraints = [
            QTextLength(QTextLength.FixedLength,    0),  # avatar
            QTextLength(QTextLength.VariableLength, 0),  # info/content
        ]

        self.msg_format = QTextTableFormat()
        self.msg_format.setBorder(0)
        self.msg_format.setTopMargin(self.font_height)
        self.msg_format.setColumnWidthConstraints(constraints)

        self.consecutive_msg_format = QTextTableFormat()
        self.consecutive_msg_format.setBorder(0)
        self.consecutive_msg_format.setColumnWidthConstraints(constraints)

        self.msg_after_break_format = QTextTableFormat()
        self.msg_after_break_format.setBorder(0)
        self.msg_after_break_format.setTopMargin(self.font_height * 3)
        self.msg_after_break_format.setColumnWidthConstraints(constraints)

        self.inner_info_content_format = QTextTableFormat()
        self.inner_info_content_format.setBorder(0)

        self.last_table_is_message: bool = False

        self.start_ms_since_epoch: int = \
            QDateTime.currentDateTime().toMSecsSinceEpoch()

        self.added_msgs: Deque[Message] = Deque()

        # [(msg.sender_id, msg.markdown)]
        self.received_by_local_echo: Deque[Tuple[str, str]] = Deque()

        uid, rid = self.chat.client.user_id, self.chat.room.room_id

        Message.local_echo_hooks[(type(self).__name__, uid, rid)] = \
            self.on_receive_local_echo

        Thread(target=self.autoload_history, daemon=True).start()

        self._add_message_request.connect(self._add_message)


    def on_receive_local_echo(self, msg: Message) -> None:
        if msg.room_id != self.chat.room.room_id:
            return

        msg.receiver_id = self.chat.client.user_id
        self.received_by_local_echo.append((msg.sender_id, msg.markdown))
        self._add_message_request.emit(msg)


    # Called from harmonyqt.chat.redirect_message()
    def on_receive_from_server(self, msg: Message) -> None:
        try:
            self.received_by_local_echo.remove((msg.sender_id, msg.markdown))
        except ValueError:  # not found in list/deque
            self._add_message_request.emit(msg)


    @staticmethod
    def break_between(msg1: Message, msg2: Message) -> bool:
        # True if there's more than 5mn between the two messages:
        return msg1.ms_since_epoch <= msg2.ms_since_epoch - 15 * 60 * 1000


    def are_consecutive(self, msg1: Message, msg2: Message) -> bool:
        return (self.last_table_is_message and
                msg1.sender_id == msg2.sender_id and
                msg1.ms_since_epoch >= msg2.ms_since_epoch - 5 * 60 * 1000)


    def _add_message(self, msg: Message) -> None:
        distance_from_left   = self.scroller.h
        distance_from_bottom = self.scroller.vmax - self.scroller.v

        cursor = QTextCursor(self.document())
        cursor.beginEditBlock()

        to_top       = msg.was_created_before(self.start_ms_since_epoch)
        inserted     = False
        previous_msg = next_msg = None

        if to_top:
            cursor.movePosition(QTextCursor.Start)

            i = 0
            for i, added_msg in enumerate(self.added_msgs, i):
                next_msg = self.added_msgs[i]

                if msg.was_created_before(added_msg.ms_since_epoch):
                    self.added_msgs.insert(i, msg)
                    cursor.movePosition(QTextCursor.Down, n=i)
                    inserted = True
                    break

                previous_msg = self.added_msgs[i - 1]

        if not inserted:
            if self.added_msgs:
                previous_msg = self.added_msgs[-1]
            self.added_msgs.append(msg)
            cursor.movePosition(QTextCursor.End)

        if next_msg and self.are_consecutive(msg, next_msg):
            fixer = QTextCursor(cursor)
            fixer.beginEditBlock()
            fixer.movePosition(QTextCursor.Down)

            msg_table = fixer.currentTable()
            if msg_table and msg_table.columns() > 1:
                msg_table.setFormat(self.consecutive_msg_format)
                fixer.select(QTextCursor.LineUnderCursor)
                fixer.removeSelectedText()  # remove avatar

            fixer.movePosition(QTextCursor.NextCell)
            fixer.movePosition(QTextCursor.NextBlock)

            info_msg_table = fixer.currentTable()
            if info_msg_table and info_msg_table.rows() > 1:
                info_msg_table.removeRows(0, 1)  # Remove info name/date

            fixer.endEditBlock()

        if next_msg and self.break_between(msg, next_msg):
            fixer = QTextCursor(cursor)
            fixer.beginEditBlock()
            fixer.movePosition(QTextCursor.Down)

            msg_table = fixer.currentTable()
            if msg_table:
                msg_table.setFormat(self.msg_after_break_format)

            fixer.endEditBlock()

        consecutive = previous_msg and self.are_consecutive(previous_msg, msg)
        after_break = previous_msg and self.break_between(previous_msg, msg)

        cursor.insertTable(
            1, 2,  # rows, columns
            self.msg_after_break_format if after_break else
            self.consecutive_msg_format if consecutive else
            self.msg_format
        )
        if not consecutive:
            cursor.insertHtml(msg.html_avatar)
        cursor.movePosition(QTextCursor.NextBlock)

        if consecutive:
            cursor.insertTable(1, 1, self.inner_info_content_format)
        else:
            cursor.insertTable(2, 1, self.inner_info_content_format)
            cursor.insertHtml(msg.html_info)
            cursor.movePosition(QTextCursor.NextBlock)

        cursor.insertHtml(msg.html_content)

        cursor.endEditBlock()
        self.last_table_is_message = True

        if to_top:
            self.scroller.hset(distance_from_left)\
                         .vset(self.scroller.vmax - distance_from_bottom)
        elif distance_from_bottom <= 10:
            self.scroller.go_min_left().go_bottom()


    def autoload_history(self) -> None:
        time.sleep(0.25)  # Give time for initial events/msgs to be shown
        scr          = self.scroller
        load_history = self.chat.room.backfill_previous_messages

        while not self.chat.room.loaded_all_history:
            if self.isVisible():

                if scr.vmax <= scr.vstep_page:
                    load_history(reverse=True, limit=25)

                elif scr.v <= scr.vmin:
                    load_history(reverse=True, limit=100)

            time.sleep(0.1)


    def keyPressEvent(self, event: QKeyEvent) -> None:
        if event.modifiers() in (Qt.NoModifier, Qt.ShiftModifier):
            self.chat.send_area.box.setFocus()
            self.chat.send_area.box.keyPressEvent(event)
            return

        super().keyPressEvent(event)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

"harmonyqt setuptools file"

from setuptools import setup, find_packages

from harmonyqt import __about__


def get_readme():
    with open("README.md", "r") as readme:
        return readme.read()


setup(
    name        = __about__.__pkg_name__,
    version     = __about__.__version__,

    author       = __about__.__author__,
    author_email = __about__.__email__,
    license      = __about__.__license__,

    description                   = __about__.__doc__,
    long_description              = get_readme(),
    long_description_content_type = "text/markdown",

    python_requires  = ">=3.6, <4",
    install_requires = [
        "PyQt5",
        "atomicfile",
        "dataclasses;python_version<'3.7'",
        "docopt",
        "matrix_client_harmonyqt[e2e]>=0.4.2-dev-4",
        "setuptools",

        "markdownify",
        "markdown2",
        "pygments",  # for markdown2 fenced code blocks coloring

        "cachetools",
        "kids.cache",
    ],


    include_package_data = True,
    packages             = find_packages(),
    package_data         = {
        __about__.__pkg_name__: ["themes/**/*", "icons/**/*"],
    },
    entry_points    = {
        "console_scripts": [
            f"{__about__.__pkg_name__}={__about__.__pkg_name__}:run",
        ]
    },

    keywords = "<KEYWORDS>",
    url      = "https://github.com/mirukan/harmonyqt",

    classifiers=[
        "Development Status :: 3 - Alpha",
        # "Development Status :: 4 - Beta",
        # "Development Status :: 5 - Production/Stable",

        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",

        "Environment :: Console",
        # "Environment :: Console :: Curses",
        # "Environment :: Plugins",
        # "Environment :: X11 Applications",
        # "Environment :: X11 Applications :: Qt",

        # "Topic :: Utilities",
        # grep '^Topic' ~/docs/web/pypi-classifiers.txt

        ("License :: OSI Approved :: "
         "GNU General Public License v3 or later (GPLv3+)"),

        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",

        "Natural Language :: English",

        "Operating System :: POSIX",
    ]
)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import re
from inspect import cleandoc
from typing import List

from . import REGISTERED_COMMANDS, register, utils
from ..chat import Chat


# pylint: disable=redefined-builtin
@register
def help(chat: Chat, args: dict) -> None:
    """Usage: /help [COMMANDS]... [-f|--full]

    Show commands overview, or the full help for COMMANDS.

    Options:
      -f, --full
        Show full help for all commands if no COMMAND is specified.

    Examples:
    ```
      /help
      /help say
      /help say nick
      /help -f
    ```"""

    help_f(chat, commands=args["COMMANDS"], full=bool(args["--full"]))


def help_f(chat: Chat, commands: List[str], full: bool = False) -> None:
    cmd_helps: List[str] = []

    full     = bool(commands) or full
    show_all = not commands
    commands = [c.lstrip("/") for c in commands] or sorted(REGISTERED_COMMANDS)

    for name in commands:
        try:
            func = REGISTERED_COMMANDS[name]
        except KeyError:
            utils.print_err(chat, f"Command not found: `{name}`.")
            continue

        if not func.__doc__:
            utils.print_err(chat, f"Missing help for `{name}`.")
            continue

        try:
            cmd_helps.append(format_doc(cleandoc(str(func.__doc__)), full))
        except Exception as err:
            utils.print_exception(chat, err, func)
            continue

    text = "<pre class=help>%s%s</pre>" % (
        "<span class='title section'>Available commands:</span><br>"
        if show_all else "",

        ("<br><br>" if full else "<br>").join(cmd_helps),
    )
    utils.print_info(chat, text, is_html=True)


class HelpParseError(Exception):
    def __init__(self, docstring: str) -> None:
        super().__init__(
            "Docstring must have an usage, short description and optionally a "
            "body (option details, etc); each separated by a blank line."
        )
        self.docstring = docstring


def format_doc(doc: str, full: bool = False) -> str:
    # Prevent < > chars in doc from being processed as HTML
    doc = doc.replace(">", "&gt;").replace("<", "&lt;")

    doc = re.sub(r"^(?!\s)(.+:)$",  # ^Title:
                 r"<span class='title command-section'>\1</span>",
                 doc,
                 flags=re.MULTILINE)
    doc = re.sub(r"(?:^|(?<=>))```\n((?:.|\n)+)\n```(?=$|<)",  # code fences
                 r"<code>\1</code>",
                 doc,
                 flags=re.MULTILINE)
    doc = re.sub(r"(?:^|(?<=[^`]))`([^`]+)`(?=$|[^`])", # `code`
                 r"<code>\1</code>",
                 doc)
    op  = {"o": r"A-Za-z\d-"}
    doc = re.sub(  # -o, --options
        r"(?:^|(?<=[^%(o)s]))(-[%(o)s]|--[%(o)s]{2,})(?=$|[^%(o)s])" % op,
        r"<code class=option>\1</code>",
        doc
    )
    doc = re.sub(  # -o/--options ARGUMENTS
        r"(<code class=option>.+?</code>\s*)([A-Z\d_]{2,})",
        r"\1<code class=option-arg>\2</code>",
        doc
    )

    try:
        usage, desc, *rest = doc.strip().split("\n\n", maxsplit=2)
    except ValueError:
        raise HelpParseError(doc)

    usage = "<code class='title command'>%s</code>" % re.sub(
        r"^(?:(?:<.+>)*usage:(?:<.+>)*)?\s*",
        "",
        usage,
        flags = re.IGNORECASE | re.MULTILINE
    )

    desc = f"<div class=description>{desc}</div>"
    body = f"<div class=body>{rest[0]}</div>" if rest else ""

    elements = [usage, desc]
    if full and body:
        elements.append(body)

    return "<div class=command-help>%s</div>" % \
           ("<br>" if full else "").join(elements)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv2.

from ..message_display import MessageDisplay


class ConsoleDisplay(MessageDisplay):
    def __init__(self) -> None:
        super().__init__()
        self.system_print_format.setTopMargin(self.font_height * 3)


    def print_error(self, trace_str: str, date: str) -> None:
        trace = trace_str.rstrip()
        self.system_print(
            f"<h3>{date}</h3><br><pre><code>{trace}</code></pre>",
            "error",
            is_html=True
        )



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

import re
import shlex
from typing import Dict

from dataclasses import dataclass

from . import eval as m_eval
from . import register, utils
from ..chat import Chat

REGISTERED_ALIASES: Dict[str, "Alias"] = {}


@dataclass
class Alias:
    alias:           str  = ""
    expands_to:      str  = ""
    single_arg:      bool = False
    is_global:       bool = False
    is_global_words: bool = False


    def __post_init__(self) -> None:
        assert self.alias and self.expands_to
        assert sum((self.single_arg, self.is_global, self.is_global_words)) < 2


    def expand(self, _: Chat, typed_text: str, force_say: bool = False) -> str:
        re_alias = re.escape(self.alias)
        re_ex    = self.expands_to.replace("\\", "\\\\")

        if self.is_global_words:
            text = re.sub(rf"(^|[^\\])({re_alias})", rf"\1{re_ex}", typed_text)
            return re.sub(rf"\\({re_alias})",        r"\1",         text)

        if self.is_global:
            text = re.sub(rf"(?:^|(?<=\s))({re_alias})(?=$|\s)",
                          re_ex, typed_text)
            return re.sub(rf"(?:^|(?<=\s))\\({re_alias})(?=$|\s)",
                          rf"\1", text)

        if force_say:
            return typed_text

        first_word, *rest = re.split(r"\s", typed_text, maxsplit=1)

        if first_word != self.alias:
            return typed_text

        if rest == []:
            return self.expands_to

        text = self.expands_to
        text = text if "{}" in text else "%s {}" % text

        if self.single_arg:
            return text.format(shlex.quote(rest[0]))

        return text.format(rest[0])


    def register(self) -> None:
        REGISTERED_ALIASES[self.alias]                    = self
        m_eval.EVAL_PARSING_HOOKS[f"alias: {self.alias}"] = self.expand

    def unregister(self) -> None:
        del REGISTERED_ALIASES[self.alias]
        del m_eval.EVAL_PARSING_HOOKS[f"alias: {self.alias}"]


@register
def alias(chat: Chat, args: dict) -> None:
    r"""
    Usage:
      /alias ALIAS
      /alias ALIAS COMMAND [-s|--single-arg|-g|--global|-G|--global-words]
      /alias ALIAS --remove

    Inspect, define or remove a shortcut for a command.

    If `COMMAND` is not specified and `ALIAS` is already defined,
    the `COMMAND` it expands to will be shown, unless `--remove` is used.

    For example, doing `/alias /h /help` will let you use `/h` as a command,
    just as if you typed `/help`.

    Note that aliases defined with this command only persist until Harmony
    is closed. To keep aliases across restarts, see `/help /autorun`.

    `ALIAS` doesn't have to start with a `/`.

    When using the created alias, arguments after it are appended as-is to
    `COMMAND`, unless `COMMAND` contains a `{}` placeholder.
    In that case, the `{}` will be replaced by the arguments.
    To have a literal `{}` in `COMMAND`, use `{{}}` instead.

    Options:
      -s, --single-arg
        Everything passed after the alias will be treated as a single argument,
        and no quoting is needed.

      -g, --global
        Allow the alias to be used anywhere in the middle of a message.
        Otherwise, the alias only works at the beginning, like normal commands.
        Global aliases do not take arguments.
        They can be escaped with a `\`, e.g. if you have a `b` global alias
        and want to say `a b c` literally, write `a \b c` instead.

      -G, --global-words
        Like --global, but also allows usage in the middle of words.

      --remove
        Remove `ALIAS` if it is defined.

    Examples:
    ```
      /alias /echo "/say --echo"
        /echo test
          Print "test" without sending it as a message

        /echo "Lorem ipsum" "sit dolor amet"
          Print "Lorem ipsum" without sending it as a message, then
          print "sit dolor amet".
          Without --single-arg, they are seen as two separate arguments.

      /alias $ "```
      $ {}
      ```"
        $ ls /tmp | wc -l
          Send the text above in a formatted markdown code block.

      /alias --single-arg /sh /shell
        /sh date
          Send the output of your OS's date command.

        /sh --echo "dir ."
          See the list of files in the current OS directory.

      /alias -s @ "/eval --user @my_other_account:matrix.org"
        @ A test message
          Say "A test message" in the chat as @my_other_account:matrix.org

        @ /nick Alice
          Change @my_other_account:matrix.org's display name to Alice

      /alias --global -> →
        Press the -> key
          Say "Press the → key" in the chat, -> is transformed to →
    ```"""

    arglias = args["ALIAS"]

    if not args["COMMAND"]:
        got = REGISTERED_ALIASES.get(arglias)

        if not got:
            utils.print_err(chat, f"No `{arglias}` alias defined.")
        elif args["--remove"]:
            got.unregister()
        else:
            utils.print_info(chat, f"`{arglias}`: `{got.expands_to}`")

        return

    Alias(alias           = arglias,
          expands_to      = args["COMMAND"],
          single_arg      = args["--single-arg"],
          is_global       = args["--global"],
          is_global_words = args["--global-words"]).register()



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv2.

from typing import Callable, Dict

from cachetools import LFUCache
from kids.cache import cache
from PyQt5.QtWidgets import QVBoxLayout, QWidget

from harmonyqt import main, main_window, message

CHAT_INIT_HOOKS: Dict[str, Callable[["Chat"], None]] = {}


class UserNotLoggedInError(Exception):
    def __init__(self, user_id: str) -> None:
        super().__init__(f"Not logged in to account {user_id!r}.")

class RoomNotJoinedError(Exception):
    def __init__(self, user_id: str, room_id: str) -> None:
        super().__init__(f"{user_id!r} is not part of the room {room_id!r}.")


@cache(use=LFUCache(maxsize=64))
class Chat(QWidget):
    def __init__(self, user_id: str, room_id: str) -> None:
        super().__init__()

        try:
            self.client = main_window().accounts[user_id]
        except KeyError:
            raise UserNotLoggedInError(user_id)

        try:
            self.room = self.client.rooms[room_id]
        except KeyError:
            raise RoomNotJoinedError(user_id, room_id)

        self.vbox = QVBoxLayout(self)
        self.vbox.setContentsMargins(0, 0, 0, 0)
        self.vbox.setSpacing(0)

        # Because of circular import
        from .display import ChatMessageDisplay
        from .send_area import SendArea
        self.messages  = ChatMessageDisplay(self)
        self.send_area = SendArea(self)

        # Make messages scroller controllable by shortcuts even when user
        # focuses sendbox, see app().get_focused parent scroller detection.
        self.scroller = self.messages.scroller

        self.vbox.addWidget(self.messages)
        self.vbox.addWidget(self.send_area)

        for hook in CHAT_INIT_HOOKS.values():
            hook(self)


def redirect_message(msg: message.Message) -> None:
    if msg.receiver_id is None:  # local echo
        return

    chat = Chat(msg.receiver_id, msg.room_id)  # cache
    chat.messages.on_receive_from_server(msg)


main.HOOKS_INIT_2_BEFORE_LOGIN["Connect chat Message redirector"] = \
    lambda win: win.events.signals.new_message.connect(redirect_message)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QLabel, QVBoxLayout, QWidget


class HomePage(QWidget):
    def __init__(self) -> None:
        super().__init__()
        self.setAttribute(Qt.WA_StyledBackground)  # Make CSS background work
        self.setMinimumWidth(10)

        label = QLabel("Placeholder home page")
        self.vbox = QVBoxLayout(self)
        self.vbox.addWidget(label, alignment=Qt.AlignCenter)



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

"""Test Matrix client"""

__pkg_name__ = "harmonyqt"
__version__  = "0.5.8"
__status__   = "Development"
# __status__ = "Production"

__author__   = "miruka"
__email__    = "miruka@disroot.org"

__license__  = "GPLv3"



# Copyright 2018 miruka
# This file is part of harmonyqt, licensed under GPLv3.

from . import base
from .login import Login
from .create_room import CreateRoom
from .accept_room_invite import AcceptRoomInvite
from .invite_to_room import InviteToRoom
